"""Tools for representing systems of units using ``astropy.units``."""
# pylint: disable=duplicate-code

__all__ = (
    "unitsystem",
    "unitsystem_of",
    # classes
    "AbstractUnitSystem",
    "UNITSYSTEMS_REGISTRY",
    "DimensionlessUnitSystem",
    "LTMAUnitSystem",
    # unit system instance
    "galactic",
    "dimensionless",
    "solarsystem",
    "si",
    "cgs",
    # unit system alias
    "NAMED_UNIT_SYSTEMS",
    # unit system flags
    "AbstractUSysFlag",
    "StandardUSysFlag",
    "DynamicalSimUSysFlag",
    # functions
    "equivalent",
)

from .base import UNITSYSTEMS_REGISTRY, AbstractUnitSystem
from .builtin import DimensionlessUnitSystem, LTMAUnitSystem
from .compare import equivalent
from .core import unitsystem, unitsystem_of
from .flags import AbstractUSysFlag, DynamicalSimUSysFlag, StandardUSysFlag
from .realizations import (
    NAMED_UNIT_SYSTEMS,
    cgs,
    dimensionless,
    galactic,
    si,
    solarsystem,
)



"""Benchmark tests for quaxed functions on quantities."""

from collections.abc import Callable
from typing import Any, TypeAlias, TypedDict

import jax
import pytest
from jax._src.stages import Compiled

import quaxed.numpy as jnp

import unxt as u

Args: TypeAlias = tuple[Any, ...]

x = jnp.linspace(0, 1, 1000)
x_nodim = u.Q(x, "")
x_length = u.Q(x, "m")
x_angle = u.Q(x, "rad")


def process_func(func: Callable[..., Any], args: Args) -> tuple[Compiled, Args]:
    """JIT and compile the function."""
    return jax.jit(func), args


class ParameterizationKWArgs(TypedDict):
    """Keyword arguments for a pytest parameterization."""

    argvalues: list[tuple[Callable[..., Any], Args]]
    ids: list[str]


def process_pytest_argvalues(
    process_fn: Callable[[Callable[..., Any], Args], tuple[Callable[..., Any], Args]],
    argvalues: list[tuple[Callable[..., Any], *tuple[Args, ...]]],
) -> ParameterizationKWArgs:
    """Process the argvalues."""
    # Get the ID for each parameterization
    get_dims = lambda args: tuple(str(u.dimension_of(a)) for a in args)
    ids: list[str] = []
    processed_argvalues: list[tuple[Compiled, Args]] = []

    for func, *many_args in argvalues:
        for args in many_args:
            ids.append(f"{func.__name__}-{get_dims(args)}")
            processed_argvalues.append(process_fn(func, args))

    # Process the argvalues and return the parameterization, with IDs
    return {"argvalues": processed_argvalues, "ids": ids}


# TODO: also benchmark BareQuantity
funcs_and_args: list[tuple[Callable[..., Any], *tuple[Args, ...]]] = [
    (jnp.abs, (x_nodim,), (x_length,)),
    (jnp.acos, (x_nodim,)),
    (jnp.acosh, (x_nodim,)),
    (jnp.add, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.asin, (x_nodim,)),
    (jnp.asinh, (x_nodim,)),
    (jnp.atan, (x_nodim,)),
    (jnp.atan2, (x_nodim, x_nodim)),
    (jnp.atanh, (x_nodim,)),
    # bitwise_and
    # bitwise_left_shift
    # bitwise_invert
    # bitwise_or
    # bitwise_right_shift
    # bitwise_xor
    (jnp.ceil, (x_nodim,), (x_length,)),
    (jnp.conj, (x_nodim,), (x_length,)),
    (jnp.cos, (x_nodim,), (x_angle,)),
    (jnp.cosh, (x_nodim,), (x_angle,)),
    (jnp.divide, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.equal, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.exp, (x_nodim,)),
    (jnp.expm1, (x_nodim,)),
    (jnp.floor, (x_nodim,), (x_length,)),
    (jnp.floor_divide, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.greater, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.greater_equal, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.imag, (x_nodim,), (x_length,)),
    (jnp.isfinite, (x_nodim,), (x_length,)),
    (jnp.isinf, (x_nodim,), (x_length,)),
    (jnp.isnan, (x_nodim,), (x_length,)),
    (jnp.less, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.less_equal, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.log, (x_nodim,)),
    (jnp.log1p, (x_nodim,)),
    (jnp.log2, (x_nodim,)),
    (jnp.log10, (x_nodim,)),
    (jnp.logaddexp, (x_nodim, x_nodim)),
    # (jnp.logical_and, (x_nodim, x_nodim)),
    # (jnp.logical_not, (x_nodim,)),
    # (jnp.logical_or, (x_nodim, x_nodim)),
    # (jnp.logical_xor, (x_nodim, x_nodim)),
    (jnp.multiply, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.negative, (x_nodim,), (x_length,)),
    (jnp.not_equal, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.positive, (x_nodim,), (x_length,)),
    # (jnp.power, (x_nodim, 2.0), (x_length, 2.0)),
    (jnp.real, (x_nodim,), (x_length,)),
    (jnp.remainder, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.round, (x_nodim,), (x_length,)),
    (jnp.sign, (x_nodim,), (x_length,)),
    (jnp.sin, (x_nodim,), (x_angle,)),
    (jnp.sinh, (x_nodim,), (x_angle,)),
    (jnp.square, (x_nodim,), (x_length,)),
    (jnp.sqrt, (x_nodim,), (x_length,)),
    (jnp.subtract, (x_nodim, x_nodim), (x_length, x_length)),
    (jnp.tan, (x_nodim,), (x_angle,)),
    (jnp.tanh, (x_nodim,), (x_angle,)),
    (jnp.trunc, (x_nodim,), (x_length,)),
]


# =============================================================================


@pytest.mark.parametrize(
    ("func", "args"), **process_pytest_argvalues(process_func, funcs_and_args)
)
@pytest.mark.benchmark(group="quaxed", max_time=1.0)
def test_jit_compile(func, args):
    """Test the speed of jitting a function."""
    _ = func.lower(*args).compile()


@pytest.mark.parametrize(
    ("func", "args"), **process_pytest_argvalues(process_func, funcs_and_args)
)
@pytest.mark.benchmark(group="quaxed", max_time=1.0)
def test_execute(func, args):
    """Test the speed of calling the function."""
    _ = jax.block_until_ready(func(*args))



"""Conversion to quantity value types."""

__all__ = ("StaticValue", "convert_to_quantity_value")

import operator
import warnings
from typing import Any, final
from typing_extensions import override

import jax
import numpy as np
import quax
import wadler_lindig as wl
from jaxtyping import Array, ArrayLike
from numpy.typing import NDArray
from plum import dispatch

import quaxed.numpy as jnp


@final
class StaticValue:
    """Immutable static value wrapper for `StaticQuantity`.

    This stores a read-only NumPy array and is used to keep the value static
    while avoiding Equinox's static-array warnings. Arithmetic operations
    degrade to the wrapped array unless both operands are `StaticValue`, in
    which case a `StaticValue` is returned.

    Note that since the array is immutable, hashing is supported.
    The hash is computed from the array's dtype, shape, and bytes.

    """

    __slots__ = ("_array",)

    def __init__(self, array: np.ndarray, /) -> None:
        value = np.asarray(array).copy()
        value.setflags(write=False)
        self._array = value

    @property
    def array(self) -> np.ndarray:
        """Return the contained NumPy array."""
        return self._array

    @property
    def _jnparray(self) -> Array:
        """Return the contained array as a JAX array."""
        return jnp.asarray(self._array)

    # Constructor
    @classmethod
    @dispatch.abstract
    def from_(cls: type["StaticValue"], *args: Any, **kwargs: Any) -> "StaticValue":
        """Create a `StaticValue` from given arguments."""
        raise NotImplementedError  # pragma: no cover

    # ==================================================
    # NumPy API

    def __array__(self, dtype: Any = None) -> np.ndarray:
        if dtype is None:
            return np.asarray(self._array)
        return np.asarray(self._array, dtype=dtype)

    # ==================================================
    # Wadler-Lindig API

    def __pdoc__(self, *, show_wrapper: bool = True, **kwargs: Any) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation of this class."""
        if not show_wrapper:
            return wl.pdoc(self._array, **kwargs)
        return (
            wl.TextDoc("StaticValue(")
            + wl.pdoc(self._array, **kwargs)
            + wl.TextDoc(")")
        )

    # ==================================================
    # Python container API

    def __len__(self) -> int:
        return len(self._array)

    def __iter__(self) -> Any:
        return iter(self._array)

    def __getitem__(self, key: Any) -> Any:
        return self._array[key]

    def __getattr__(self, name: str) -> Any:
        return getattr(jnp.asarray(self._array), name)

    def __repr__(self) -> str:
        return wl.pformat(self, short_arrays=False, max_width=80)

    @override
    def __eq__(self, other: object, /) -> bool | NDArray[np.bool_]:  # type: ignore[override]
        if isinstance(other, StaticValue):
            return bool(np.array_equal(self._array, other._array))
        if isinstance(other, (np.ndarray, Array, list, tuple)):
            return self._array == other
        return NotImplemented

    def __ne__(self, other: object, /) -> bool | NDArray[np.bool_]:  # type: ignore[override]
        if isinstance(other, StaticValue):
            return not bool(np.array_equal(self._array, other._array))
        if isinstance(other, (np.ndarray, Array, list, tuple)):
            return self._array != other
        return NotImplemented

    def __lt__(self, other: Any, /) -> NDArray[np.bool_]:
        if isinstance(other, StaticValue):
            other = other._array
        return self._array < other

    def __le__(self, other: Any, /) -> NDArray[np.bool_]:
        if isinstance(other, StaticValue):
            other = other._array
        return self._array <= other

    def __gt__(self, other: Any, /) -> NDArray[np.bool_]:
        if isinstance(other, StaticValue):
            other = other._array
        return self._array > other

    def __ge__(self, other: Any, /) -> NDArray[np.bool_]:
        if isinstance(other, StaticValue):
            other = other._array
        return self._array >= other

    def __hash__(self) -> int:
        return hash((self._array.dtype.str, self._array.shape, self._array.tobytes()))

    def _binary_op(self, other: Any, op: Any) -> Any:
        if isinstance(other, StaticValue):
            result = op(self._jnparray, jnp.asarray(other.array))
            return StaticValue(np.asarray(result))
        return op(self._jnparray, other)

    def _rbinary_op(self, other: Any, op: Any) -> Any:
        if isinstance(other, StaticValue):
            result = op(jnp.asarray(other.array), self._jnparray)
            return StaticValue(np.asarray(result))
        return op(other, self._jnparray)

    def __add__(self, other: Any) -> Any:
        return self._binary_op(other, operator.add)

    def __radd__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.add)

    def __sub__(self, other: Any) -> Any:
        return self._binary_op(other, operator.sub)

    def __rsub__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.sub)

    def __mul__(self, other: Any) -> Any:
        return self._binary_op(other, operator.mul)

    def __rmul__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.mul)

    def __truediv__(self, other: Any) -> Any:
        return self._binary_op(other, operator.truediv)

    def __rtruediv__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.truediv)

    def __floordiv__(self, other: Any) -> Any:
        return self._binary_op(other, operator.floordiv)

    def __rfloordiv__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.floordiv)

    def __pow__(self, other: Any) -> Any:
        return self._binary_op(other, operator.pow)

    def __rpow__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.pow)

    def __mod__(self, other: Any) -> Any:
        return self._binary_op(other, operator.mod)

    def __rmod__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.mod)

    def __matmul__(self, other: Any) -> Any:
        return self._binary_op(other, operator.matmul)

    def __rmatmul__(self, other: Any) -> Any:
        return self._rbinary_op(other, operator.matmul)

    def __neg__(self) -> Any:
        return -self._jnparray

    def __pos__(self) -> Any:
        return +self._jnparray

    def __abs__(self) -> Any:
        return abs(self._jnparray)


@StaticValue.from_.dispatch
def from_(cls: type[StaticValue], value: object, /) -> StaticValue:
    """Convert a value for `StaticQuantity`."""
    return cls(np.asarray(value))


@StaticValue.from_.dispatch
def from_(cls: type[StaticValue], value: StaticValue, /) -> StaticValue:
    """Convert a value for `StaticQuantity`."""
    return value


@StaticValue.from_.dispatch
def from_(cls: type[StaticValue], value: jax.Array | jax.core.Tracer, /) -> StaticValue:
    """Reject JAX arrays for `StaticQuantity`."""
    msg = "StaticQuantity does not accept JAX arrays. Use Quantity for traced values."
    raise TypeError(msg)


# ==================================================================
# Converters for quantity value field


@dispatch.abstract
def convert_to_quantity_value(obj: Any, /) -> Any:
    """Convert for the value field of an `AbstractQuantity` subclass."""
    raise NotImplementedError  # pragma: no cover


@dispatch
def convert_to_quantity_value(obj: StaticValue, /) -> StaticValue:
    """Allow static values in `Quantity`-like classes."""
    return obj


@dispatch
def convert_to_quantity_value(obj: quax.ArrayValue, /) -> Any:
    """Convert a `quax.ArrayValue` for the value field.

    >>> import warnings
    >>> import jax
    >>> import jax.numpy as jnp
    >>> from jaxtyping import Array
    >>> from quax import ArrayValue

    >>> class MyArray(ArrayValue):
    ...     value: Array
    ...
    ...     def aval(self):
    ...         return jax.core.ShapedArray(self.value.shape, self.value.dtype)
    ...
    ...     def materialise(self):
    ...         return self.value

    >>> x = MyArray(jnp.array([1, 2, 3]))
    >>> with warnings.catch_warnings(record=True, action="always") as w:
    ...     y = convert_to_quantity_value(x)
    >>> y
    MyArray(value=i32[3])
    >>> print(f"Warning caught: {w[-1].message}")
    Warning caught: 'quax.ArrayValue' subclass 'MyArray' ...

    """
    warnings.warn(
        f"'quax.ArrayValue' subclass {type(obj).__name__!r} does not have a registered "
        "converter. Returning the object as is.",
        category=UserWarning,
        stacklevel=2,
    )
    return obj


@dispatch
def convert_to_quantity_value(obj: ArrayLike | list[Any] | tuple[Any, ...], /) -> Array:
    """Convert an array-like object to a `jax.numpy.ndarray`.

    >>> import jax.numpy as jnp
    >>> from unxt.quantity import convert_to_quantity_value

    >>> convert_to_quantity_value([1, 2, 3])
    Array([1, 2, 3], dtype=int32)

    """
    return jnp.asarray(obj)



# pylint: disable=import-error, no-member, unsubscriptable-object
#    b/c it doesn't understand dataclass fields

__all__ = ("BareQuantity",)

from typing import Any

import equinox as eqx
from jaxtyping import Array, Shaped
from plum import add_promotion_rule

from .base import AbstractQuantity
from .quantity import Quantity
from .value import StaticValue, convert_to_quantity_value
from unxt.units import AbstractUnit, unit as parse_unit


class BareQuantity(AbstractQuantity):
    """A fast implementation of the Quantity class.

    This class is not parametrized by its dimensionality.
    """

    value: Shaped[Array | StaticValue, "*shape"] = eqx.field(
        converter=convert_to_quantity_value
    )
    """The value of the `AbstractQuantity`."""

    unit: AbstractUnit = eqx.field(static=True, converter=parse_unit)
    """The unit associated with this value."""

    def __class_getitem__(cls: "type[BareQuantity]", item: Any) -> "type[BareQuantity]":
        """No-op support for `BareQuantity[...]` syntax.

        This method is called when the class is subscripted, e.g.:

        >>> from unxt.quantity import BareQuantity
        >>> BareQuantity["length"]
        <class 'unxt...quantity...BareQuantity'>

        """
        return cls


add_promotion_rule(BareQuantity, Quantity, Quantity)



"""Unit tests for quantity API (unxt_api._quantity)."""

import inspect
from typing import Any

import plum
import pytest

import unxt_api as uapi

# ==============================================================================
# Tests for uconvert_value()
# ==============================================================================


def test_uconvert_value_is_abstract_dispatch() -> None:
    """Test that uconvert_value is an abstract dispatch function."""
    assert isinstance(uapi.uconvert_value, plum.function.Function)
    assert hasattr(uapi.uconvert_value, "methods")


def test_uconvert_value_accepts_any_type() -> None:
    """Test that uconvert_value accepts Any types."""
    assert "uconvert_value" in dir(uapi)


def test_uconvert_value_no_default_implementation_raises() -> None:
    """Test that calling uconvert_value without implementation raises error."""

    class NoDispatchType:
        pass

    obj1 = NoDispatchType()
    obj2 = NoDispatchType()
    obj3 = NoDispatchType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.uconvert_value(obj1, obj2, obj3)


def test_uconvert_value_can_register_custom_dispatch(custom_unit_type) -> None:
    """Test that custom dispatches can be registered."""

    @plum.dispatch
    def uconvert_value(
        uto: custom_unit_type, ufrom: custom_unit_type, x: float, /
    ) -> float:
        # Simple mock conversion (1 km = 1000 m)
        conversion_factor = (
            1000.0 if ufrom.unit_str == "km" and uto.unit_str == "m" else 1.0
        )
        return x * conversion_factor

    from_unit = custom_unit_type("km")
    to_unit = custom_unit_type("m")
    result = uconvert_value(to_unit, from_unit, 5.0)

    assert result == 5000.0


def test_uconvert_value_signature_order() -> None:
    """Test uconvert_value parameter order (to_unit, from_unit, value)."""
    # This validates the documented order: (uto, ufrom, x)
    # The docstring example shows this clearly:
    # u.uconvert_value(u.unit("m"), u.unit("km"), 1)  # (to, from, value)
    sig = inspect.signature(uapi.uconvert_value.__wrapped__)
    params = list(sig.parameters.keys())
    assert params == ["uto", "ufrom", "x"]


# ==============================================================================
# Tests for uconvert()
# ==============================================================================


def test_uconvert_is_abstract_dispatch() -> None:
    """Test that uconvert is an abstract dispatch function."""
    assert isinstance(uapi.uconvert, plum.function.Function)
    assert hasattr(uapi.uconvert, "methods")


def test_uconvert_accepts_any_type() -> None:
    """Test that uconvert accepts Any type in signature."""
    assert "uconvert" in dir(uapi)


def test_uconvert_no_default_implementation_raises() -> None:
    """Test that calling uconvert without implementation raises error."""

    class NoDispatchType:
        pass

    obj1 = NoDispatchType()
    obj2 = NoDispatchType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.uconvert(obj1, obj2)


def test_uconvert_can_register_custom_dispatch(
    custom_unit_type, custom_quantity_type
) -> None:
    """Test that custom dispatches can be registered."""

    @plum.dispatch
    def uconvert(
        to_unit: custom_unit_type, quantity: custom_quantity_type, /
    ) -> custom_quantity_type:
        # Simple mock conversion
        return custom_quantity_type(quantity.value, to_unit.unit_str)

    unit = custom_unit_type("m")
    quantity = custom_quantity_type(1000, "km")
    result = uconvert(unit, quantity)

    assert isinstance(result, custom_quantity_type)
    assert result.unit_str == "m"


# ==============================================================================
# Tests for ustrip()
# ==============================================================================


def test_ustrip_is_abstract_dispatch() -> None:
    """Test that ustrip is an abstract dispatch function."""
    assert isinstance(uapi.ustrip, plum.function.Function)
    assert hasattr(uapi.ustrip, "methods")


def test_ustrip_accepts_any_type() -> None:
    """Test that ustrip accepts varargs."""
    assert "ustrip" in dir(uapi)


def test_ustrip_no_default_implementation_raises() -> None:
    """Test that calling ustrip without implementation raises error."""

    class NoDispatchType:
        pass

    obj = NoDispatchType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.ustrip(obj)


def test_ustrip_can_register_custom_dispatch(custom_quantity_type) -> None:
    """Test that custom dispatches can be registered for ustrip."""

    @plum.dispatch
    def ustrip(quantity: custom_quantity_type, /) -> Any:
        return quantity.value

    quantity = custom_quantity_type(42.5, "m")
    result = ustrip(quantity)
    assert result == 42.5


def test_ustrip_varargs_signature() -> None:
    """Test that ustrip supports variable arguments."""
    # The signature should accept *args
    # This is reflected in the abstract signature


# ==============================================================================
# Tests for is_unit_convertible()
# ==============================================================================


def test_is_unit_convertible_is_abstract_dispatch() -> None:
    """Test that is_unit_convertible is an abstract dispatch function."""
    assert isinstance(uapi.is_unit_convertible, plum.function.Function)
    assert hasattr(uapi.is_unit_convertible, "methods")


def test_is_unit_convertible_accepts_any_type() -> None:
    """Test that is_unit_convertible accepts Any types."""
    assert "is_unit_convertible" in dir(uapi)


def test_is_unit_convertible_no_default_implementation() -> None:
    """Test that is_unit_convertible raises when no dispatch found."""

    class NoDispatchType:
        pass

    obj1 = NoDispatchType()
    obj2 = NoDispatchType()

    # Should raise NotFoundLookupError when no dispatch is registered
    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.is_unit_convertible(obj1, obj2)


def test_is_unit_convertible_can_register_custom_dispatch(custom_unit_type) -> None:
    """Test that custom dispatches can be registered."""

    @plum.dispatch
    def is_unit_convertible(
        to_unit: custom_unit_type, from_unit: custom_unit_type, /
    ) -> bool:
        # Simple mock: same dimension means convertible
        return to_unit.unit_str[0] == from_unit.unit_str[0]

    unit1 = custom_unit_type("m")
    unit2 = custom_unit_type("mm")
    unit3 = custom_unit_type("kg")

    assert is_unit_convertible(unit1, unit2) is True
    assert is_unit_convertible(unit1, unit3) is False


def test_is_unit_convertible_returns_bool() -> None:
    """Test that return type annotation is bool."""
    # The abstract signature returns bool
    inspect.signature(uapi.is_unit_convertible.__wrapped__)
    # Note: __wrapped__ may not be available, but we can check the docstring


# ==============================================================================
# Tests for wrap_to()
# ==============================================================================


def test_wrap_to_is_abstract_dispatch() -> None:
    """Test that wrap_to is an abstract dispatch function."""
    assert isinstance(uapi.wrap_to, plum.function.Function)
    assert hasattr(uapi.wrap_to, "methods")


def test_wrap_to_accepts_any_type() -> None:
    """Test that wrap_to accepts Any types."""
    assert "wrap_to" in dir(uapi)


def test_wrap_to_no_default_implementation_raises() -> None:
    """Test that calling wrap_to without implementation raises error."""

    class NoDispatchType:
        pass

    obj = NoDispatchType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.wrap_to(obj, obj, obj)


def test_wrap_to_can_register_custom_dispatch(custom_quantity_type) -> None:
    """Test that custom dispatches can be registered."""

    @plum.dispatch
    def wrap_to(
        x: custom_quantity_type,
        min: custom_quantity_type,
        max: custom_quantity_type,
        /,
    ) -> custom_quantity_type:
        # Simple wrapping logic
        range_size = max.value - min.value
        wrapped = ((x.value - min.value) % range_size) + min.value
        return custom_quantity_type(wrapped, x.unit_str)

    x = custom_quantity_type(370, "deg")
    min_val = custom_quantity_type(0, "deg")
    max_val = custom_quantity_type(360, "deg")

    result = wrap_to(x, min_val, max_val)
    assert result.value == 10


def test_wrap_to_keyword_argument_dispatch() -> None:
    """Test that wrap_to supports keyword arguments through redirect."""
    # The package provides a keyword argument dispatch that redirects
    # to the positional version
    # Check that both dispatches exist
    assert len(uapi.wrap_to.methods) >= 2


# ==============================================================================
# API consistency tests
# ==============================================================================


def test_all_quantity_functions_are_exported() -> None:
    """Test that all quantity functions are exported."""
    assert hasattr(uapi, "uconvert_value")
    assert hasattr(uapi, "uconvert")
    assert hasattr(uapi, "ustrip")
    assert hasattr(uapi, "is_unit_convertible")
    assert hasattr(uapi, "wrap_to")


def test_all_quantity_functions_in_all() -> None:
    """Test that all quantity functions are in __all__."""
    assert "uconvert_value" in uapi.__all__
    assert "uconvert" in uapi.__all__
    assert "ustrip" in uapi.__all__
    assert "is_unit_convertible" in uapi.__all__
    assert "wrap_to" in uapi.__all__


def test_all_quantity_functions_are_dispatch_functions() -> None:
    """Test that all quantity functions are dispatch functions."""
    assert isinstance(uapi.uconvert_value, plum.function.Function)
    assert isinstance(uapi.uconvert, plum.function.Function)
    assert isinstance(uapi.ustrip, plum.function.Function)
    assert isinstance(uapi.is_unit_convertible, plum.function.Function)
    assert isinstance(uapi.wrap_to, plum.function.Function)



"""Tests for st.from_type() with registered type strategies."""

import hypothesis.strategies as st
import pytest
from hypothesis import HealthCheck, given, settings

import unxt as u
import unxt_hypothesis as ust  # noqa: F401 - Import to trigger type strategy registrations


@given(q=st.from_type(u.AbstractQuantity))
@settings(max_examples=20)
def test_from_type_abstract_quantity(q: u.AbstractQuantity) -> None:
    """Test st.from_type() with AbstractQuantity."""
    assert isinstance(q, u.AbstractQuantity)
    # Should have a dimension
    assert u.dimension_of(q) is not None


@given(a=st.from_type(u.Angle))
@settings(max_examples=20)
def test_from_type_angle(a: u.Angle) -> None:
    """Test st.from_type() with Angle."""
    assert isinstance(a, u.Angle)
    # Should have angle dimension
    assert u.dimension_of(a) == u.dimension("angle")


@given(q=st.from_type(u.Quantity))
@settings(max_examples=20)
def test_from_type_quantity(q: u.Quantity) -> None:
    """Test st.from_type() with Quantity."""
    assert isinstance(q, u.Quantity)
    # Should have a dimension
    assert u.dimension_of(q) is not None


@given(bq=st.from_type(u.quantity.BareQuantity))
@settings(max_examples=20)
def test_from_type_bare_quantity(bq: u.quantity.BareQuantity) -> None:
    """Test st.from_type() with BareQuantity."""
    assert isinstance(bq, u.quantity.BareQuantity)
    # BareQuantity should still have a unit
    assert bq.unit is not None


@given(sq=st.from_type(u.quantity.StaticQuantity))
@settings(max_examples=20)
def test_from_type_static_quantity(sq: u.quantity.StaticQuantity) -> None:
    """Test st.from_type() with StaticQuantity."""
    assert isinstance(sq, u.quantity.StaticQuantity)
    # StaticQuantity should have a StaticValue wrapper
    assert isinstance(sq.value, u.quantity.StaticValue)
    # Should still have a unit
    assert sq.unit is not None


@pytest.mark.filterwarnings(
    "ignore:Do not use the `random` module inside "
    "strategies:hypothesis.errors.HypothesisDeprecationWarning"
)
@given(us=st.from_type(u.AbstractUnitSystem))
@settings(
    max_examples=20,
    suppress_health_check=[HealthCheck.differing_executors],
)
def test_from_type_abstract_unitsystem(us: u.AbstractUnitSystem) -> None:
    """Test st.from_type() with AbstractUnitSystem."""
    assert isinstance(us, u.AbstractUnitSystem)
    # Unit systems can be empty or populated
    assert len(us) >= 0



"""Register dispatches for `quaxed.numpy`."""
# pylint: disable=import-error

__all__: tuple[str, ...] = ()

from typing import Any

import jax
import jax.numpy as jax_xp
from jaxtyping import ArrayLike
from plum import dispatch
from plum.parametric import type_unparametrized as type_np

from .base import AbstractQuantity
from .quantity import Quantity
from unxt_api import ustrip

# -----------------------------------------------


@dispatch
def arange(
    start: AbstractQuantity,
    stop: AbstractQuantity | None = None,
    step: AbstractQuantity | None = None,
    **kwargs: Any,
) -> AbstractQuantity:
    """Return evenly spaced values within a given interval.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.arange(Quantity(5, "m"))
    Quantity(Array([0, 1, 2, 3, 4], dtype=int32), unit='m')

    >>> jnp.arange(Quantity(5, "m"), Quantity(10, "m"))
    Quantity(Array([5, 6, 7, 8, 9], dtype=int32), unit='m')

    >>> jnp.arange(Quantity(5, "m"), Quantity(10, "m"), Quantity(2, "m"))
    Quantity(Array([5, 7, 9], dtype=int32), unit='m')

    """
    unit = start.unit
    return type_np(start)(
        jax_xp.arange(
            start.value,
            stop=ustrip(unit, stop) if stop is not None else None,
            step=ustrip(unit, step) if step is not None else None,
            **kwargs,
        ),
        unit=unit,
    )


# -----------------------------------------------


@dispatch
def empty_like(
    x: AbstractQuantity, /, *, device: Any = None, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array with the same shape and type as a given array.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.empty_like(Quantity(5, "m"))
    Quantity(Array(0, dtype=int32, ...), unit='m')

    """
    out = type_np(x)(jax_xp.empty_like(x.value, **kwargs), unit=x.unit)
    return jax.device_put(out, device=device)


# -----------------------------------------------


@dispatch
def full(
    shape: Any, fill_value: AbstractQuantity, /, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array of given shape and type, filled with `fill_value`.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.full((2, 2), Quantity(5, "m"))
    Quantity(Array([[5, 5], [5, 5]], dtype=int32, ...), unit='m')

    """
    fill_val = ustrip(fill_value.unit, fill_value)
    return Quantity(jax_xp.full(shape, fill_val, **kwargs), unit=fill_value.unit)


# -----------------------------------------------


@dispatch
def full_like(
    x: AbstractQuantity, /, *, fill_value: Any, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array with the same shape and type as a given array.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.full_like(Quantity(5, "m"), fill_value=Quantity(10, "m"))
    Quantity(Array(10, dtype=int32, ...), unit='m')

    """
    # re-dispatch to the correct implementation
    return full_like(x, fill_value, **kwargs)


@dispatch
def full_like(
    x: AbstractQuantity, fill_value: ArrayLike, /, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array with the same shape and type as a given array.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.full_like(Quantity(5, "m"), 100.0)
    Quantity(Array(100, dtype=int32, ...), unit='m')

    """
    return type_np(x)(jax_xp.full_like(x.value, fill_value, **kwargs), unit=x.unit)


@dispatch
def full_like(
    x: AbstractQuantity, fill_value: AbstractQuantity, /, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array with the same shape and type as a given array.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.full_like(Quantity(5, "m"), Quantity(10, "m"))
    Quantity(Array(10, dtype=int32, ...), unit='m')

    """
    fill_val = ustrip(x.unit, fill_value)
    return type_np(x)(jax_xp.full_like(x.value, fill_val, **kwargs), unit=x.unit)


# -----------------------------------------------


@dispatch
def linspace(
    start: AbstractQuantity, stop: AbstractQuantity, num: Any, /, **kwargs: Any
) -> AbstractQuantity:
    """Return evenly spaced values within a given interval.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.linspace(Quantity(0, "m"), Quantity(10, "m"), 5)
    Quantity(Array([ 0. ,  2.5,  5. ,  7.5, 10. ], dtype=float32), unit='m')

    """
    unit = start.unit
    return type_np(start)(
        jax_xp.linspace(ustrip(unit, start), ustrip(unit, stop), num, **kwargs),
        unit=unit,
    )


@dispatch
def linspace(
    start: AbstractQuantity, stop: AbstractQuantity, /, **kwargs: Any
) -> AbstractQuantity:
    """Return evenly spaced values within a given interval.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.linspace(Quantity(0, "m"), Quantity(10, "m"), num=5)
    Quantity(Array([ 0. ,  2.5,  5. ,  7.5, 10. ], dtype=float32), unit='m')

    """
    unit = start.unit
    return type_np(start)(
        jax_xp.linspace(ustrip(unit, start), ustrip(unit, stop), **kwargs), unit=unit
    )


# -----------------------------------------------


@dispatch
def ones_like(
    x: AbstractQuantity, /, *, device: Any = None, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array with the same shape and type as a given array.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.ones_like(Quantity(5, "m"))
    Quantity(Array(1, dtype=int32, ...), unit='m')

    """
    out = type_np(x)(jax_xp.ones_like(x.value, **kwargs), unit=x.unit)
    return jax.device_put(out, device=device)


# -----------------------------------------------


@dispatch
def zeros_like(
    x: AbstractQuantity, /, *, device: Any = None, **kwargs: Any
) -> AbstractQuantity:
    """Return a new array with the same shape and type as a given array.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity

    >>> jnp.zeros_like(Quantity(5, "m"))
    Quantity(Array(0, dtype=int32, ...), unit='m')

    """
    out = type_np(x)(jax_xp.zeros_like(x.value, **kwargs), unit=x.unit)
    return jax.device_put(out, device=device)



"""Tests for StaticQuantity."""

from functools import partial

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import wadler_lindig as wl
from hypothesis import given, strategies as st
from hypothesis.extra.numpy import arrays as np_arrays
from plum import promote

import unxt as u


def test_static_quantity_accepts_numpy() -> None:
    """StaticQuantity accepts Python scalars and NumPy arrays."""
    scalar = u.StaticQuantity(1, "m")
    assert np.array_equal(np.asarray(scalar.value), np.asarray(1))

    arr = np.array([1.0, 2.0])
    vec = u.StaticQuantity(arr, "m")
    assert np.array_equal(np.asarray(vec.value), arr)


def test_static_quantity_rejects_jax_array() -> None:
    """StaticQuantity rejects JAX arrays."""
    with pytest.raises(TypeError, match="StaticQuantity does not accept JAX arrays"):
        u.StaticQuantity(jnp.array([1.0, 2.0]), "m")


def test_static_quantity_accepts_array_like() -> None:
    """StaticQuantity accepts array-like inputs."""
    q = u.StaticQuantity([1.0, 2.0], "m")
    assert np.array_equal(np.asarray(q.value), np.array([1.0, 2.0]))


def test_static_quantity_hashable_python() -> None:
    """StaticQuantity can be hashed in Python."""
    q1 = u.StaticQuantity(1, "m")
    q2 = u.StaticQuantity(1, "m")
    assert hash(q1) == hash(q2)

    q_arr1 = u.StaticQuantity(np.array([1.0, 2.0]), "m")
    q_arr2 = u.StaticQuantity(np.array([1.0, 2.0]), "m")
    assert hash(q_arr1) == hash(q_arr2)


def test_static_quantity_hashable_jax() -> None:
    """StaticQuantity can be used as a static arg in jitted code."""
    q = u.StaticQuantity(np.array([1.0, 2.0]), "m")

    @partial(jax.jit, static_argnames=("q",))
    def add(x, q):
        return x + jnp.asarray(q.value)

    out = add(1.0, q)
    assert np.allclose(out, np.array([2.0, 3.0]))

    out_again = add(1.0, u.StaticQuantity(np.array([1.0, 2.0]), "m"))
    assert np.allclose(out_again, np.array([2.0, 3.0]))


def test_static_quantity_vmap_static_arg() -> None:
    """StaticQuantity can be used as a static arg in vmapped code."""
    q = u.StaticQuantity(np.array([1.0, 2.0]), "m")

    def add(x, q):
        return x + jnp.asarray(q.value)

    out = jax.vmap(add, in_axes=(0, None))(jnp.array([1.0, 2.0]), q)
    assert np.allclose(out, np.array([[2.0, 3.0], [3.0, 4.0]]))


def test_static_quantity_promotes_to_quantity() -> None:
    """StaticQuantity promotes to Quantity."""
    q = u.StaticQuantity(np.array([1.0, 2.0]), "m")
    r = u.Q(jnp.array([3.0, 4.0]), "m")

    q_promoted, r_promoted = promote(q, r)
    assert isinstance(q_promoted, u.Q)
    assert isinstance(r_promoted, u.Q)


def test_static_quantity_subtraction_with_quantity() -> None:
    """StaticQuantity subtraction with Quantity promotes to Quantity."""
    sq = u.StaticQuantity(np.array(1.0), "s")
    q = u.Quantity(0.5, "s")

    # Quantity - StaticQuantity -> Quantity
    result1 = q - sq
    assert isinstance(result1, u.Q)
    assert np.allclose(np.asarray(result1.value), -0.5)

    # StaticQuantity - Quantity -> Quantity
    result2 = sq - q
    assert isinstance(result2, u.Q)
    assert np.allclose(np.asarray(result2.value), 0.5)


def test_static_quantity_subtraction_preserves_static() -> None:
    """StaticQuantity subtraction with StaticQuantity stays StaticQuantity."""
    sq1 = u.StaticQuantity(np.array(1.0), "km")
    sq2 = u.StaticQuantity(np.array(500.0), "m")

    result = sq1 - sq2
    assert isinstance(result, u.StaticQuantity)
    assert np.allclose(np.asarray(result.value), 0.5)
    assert result.unit == sq1.unit


def test_static_quantity_addition_with_quantity() -> None:
    """StaticQuantity addition with Quantity promotes to Quantity."""
    sq = u.StaticQuantity(np.array(1.0), "s")
    q = u.Quantity(0.5, "s")

    # Quantity + StaticQuantity -> Quantity
    result1 = q + sq
    assert isinstance(result1, u.Q)
    assert np.allclose(np.asarray(result1.value), 1.5)

    # StaticQuantity + Quantity -> Quantity
    result2 = sq + q
    assert isinstance(result2, u.Q)
    assert np.allclose(np.asarray(result2.value), 1.5)


def test_static_quantity_addition_preserves_static() -> None:
    """StaticQuantity addition with StaticQuantity stays StaticQuantity."""
    sq1 = u.StaticQuantity(np.array(1.0), "km")
    sq2 = u.StaticQuantity(np.array(500.0), "m")

    result = sq1 + sq2
    assert isinstance(result, u.StaticQuantity)
    assert np.allclose(np.asarray(result.value), 1.5)
    assert result.unit == sq1.unit


def test_static_quantity_multiplication_with_quantity() -> None:
    """StaticQuantity multiplication with Quantity promotes to Quantity."""
    sq = u.StaticQuantity(2.0, "m")
    q = u.Quantity(3.0, "s")

    # Quantity * StaticQuantity -> Quantity
    result1 = q * sq
    assert isinstance(result1, u.Q)
    assert np.allclose(result1.value, 6.0)
    assert result1.unit == u.unit("m * s")

    # StaticQuantity * Quantity -> Quantity
    result2 = sq * q
    assert isinstance(result2, u.Q)
    assert np.allclose(result2.value, 6.0)
    assert result2.unit == u.unit("m * s")


def test_static_quantity_multiplication_preserves_static() -> None:
    """StaticQuantity multiplication with StaticQuantity stays StaticQuantity."""
    sq1 = u.StaticQuantity(2.0, "m")
    sq2 = u.StaticQuantity(3.0, "s")

    result = sq1 * sq2
    assert isinstance(result, u.StaticQuantity)
    assert np.allclose(result.value, 6.0)
    assert result.unit == u.unit("m * s")


def test_static_quantity_division_with_quantity() -> None:
    """StaticQuantity division with Quantity promotes to Quantity."""
    sq = u.StaticQuantity(6.0, "m")
    q = u.Quantity(2.0, "s")

    # Quantity / StaticQuantity -> Quantity
    result1 = q / sq
    assert isinstance(result1, u.Q)
    assert np.allclose(result1.value, 1.0 / 3.0)
    assert result1.unit == u.unit("s / m")

    # StaticQuantity / Quantity -> Quantity
    result2 = sq / q
    assert isinstance(result2, u.Q)
    assert np.allclose(result2.value, 3.0)
    assert result2.unit == u.unit("m / s")


def test_static_quantity_division_preserves_static() -> None:
    """StaticQuantity division with StaticQuantity stays StaticQuantity."""
    sq1 = u.StaticQuantity(6.0, "m")
    sq2 = u.StaticQuantity(2.0, "s")

    result = sq1 / sq2
    assert isinstance(result, u.StaticQuantity)
    assert np.allclose(result.value, 3.0)
    assert result.unit == u.unit("m / s")


def test_static_quantity_division_integer_inputs() -> None:
    """StaticQuantity division with integer inputs promotes to float."""
    sq1 = u.StaticQuantity(6, "m")
    sq2 = u.StaticQuantity(2, "s")

    result = sq1 / sq2
    assert isinstance(result, u.StaticQuantity)
    # Integer division should promote to float (true division semantics)
    assert result.value.dtype == np.float32
    assert np.allclose(result.value, 3.0)
    assert result.unit == u.unit("m / s")


def test_static_quantity_modulo_with_quantity() -> None:
    """StaticQuantity modulo with Quantity promotes to Quantity."""
    sq = u.StaticQuantity(7.0, "m")
    q = u.Quantity(3.0, "m")

    # Quantity % StaticQuantity -> Quantity (via promotion)
    result1 = q % sq
    assert isinstance(result1, u.Q)
    assert np.allclose(result1.value, 3.0 % 7.0)
    assert result1.unit == u.unit("m")

    # StaticQuantity % Quantity -> Quantity (via promotion)
    result2 = sq % q
    assert isinstance(result2, u.Q)
    assert np.allclose(result2.value, 7.0 % 3.0)
    assert result2.unit == u.unit("m")


def test_static_quantity_modulo_with_static_quantity_promotes() -> None:
    """StaticQuantity modulo StaticQuantity promotes to Quantity."""
    sq1 = u.StaticQuantity(7.0, "m")
    sq2 = u.StaticQuantity(3.0, "m")

    result = sq1 % sq2
    # Since there's no modulo_p primitive dispatch, plum promotion rules apply
    # StaticQuantity % StaticQuantity -> Quantity (via default promotion)
    assert isinstance(result, u.Q)
    assert np.allclose(result.value, 7.0 % 3.0)
    assert result.unit == u.unit("m")


def test_static_quantity_modulo_preserves_numpy_input_values() -> None:
    """StaticQuantity stores numpy arrays that persist through operations."""
    # Verify that StaticQuantity with numpy arrays stores them correctly
    sq = u.StaticQuantity(np.array([7.0, 8.0, 9.0]), "m")

    # The underlying value should be a StaticValue wrapping a numpy array
    assert isinstance(sq.value, u.quantity.StaticValue)
    # Verify the array inside is numpy, not JAX
    underlying_array = sq.value.array
    assert isinstance(underlying_array, np.ndarray)
    assert not isinstance(underlying_array, jnp.ndarray)
    assert np.array_equal(underlying_array, np.array([7.0, 8.0, 9.0]))


def test_static_value_modulo_operations() -> None:
    """StaticValue modulo operations work correctly with forward and reverse."""
    sv = u.quantity.StaticValue(np.array([7.0, 8.0]))

    # Forward modulo
    result_forward = sv % 3.0
    assert np.allclose(result_forward, np.array([1.0, 2.0]))

    # Reverse modulo
    result_reverse = 10.0 % sv
    assert np.allclose(result_reverse, np.array([3.0, 2.0]))
    # With another StaticValue
    sv2 = u.quantity.StaticValue(np.array([3.0, 3.0]))
    result_sv = sv % sv2
    assert isinstance(result_sv, u.quantity.StaticValue)
    assert np.allclose(result_sv.array, np.array([1.0, 2.0]))


def test_static_value_pdoc() -> None:
    """StaticValue uses the contained value for formatting."""
    value = u.quantity.StaticValue(np.array([1.0, 2.0]))
    assert wl.pformat(value, show_wrapper=False) == wl.pformat(np.array([1.0, 2.0]))


def test_static_quantity_pdoc_hides_static_value() -> None:
    """StaticQuantity formatting should not expose StaticValue."""
    q = u.StaticQuantity(np.array([1.0, 2.0]), "m")
    formatted = wl.pformat(q, short_arrays=False)
    assert "StaticValue" not in formatted
    assert "unit='m'" in formatted


def test_quantity_static_value_jit_grad() -> None:
    """StaticValue should behave as static in JAX transforms."""
    sv = u.quantity.StaticValue(np.array([1.0, 2.0]))
    q = u.Q(sv, "m")

    def f(x, q):
        return jnp.sum(jnp.asarray(q.value) * x)

    jit_f = eqx.filter_jit(f)
    out = jit_f(jnp.array([1.0, 1.0]), q)
    assert np.allclose(out, np.array(3.0))

    grad_f = eqx.filter_grad(lambda x, q: f(x, q))
    grad = grad_f(jnp.array([1.0, 1.0]), q)
    assert np.allclose(grad, np.array([1.0, 2.0]))


def test_static_value_binary_ops_degrade() -> None:
    """StaticValue should behave like its array in arithmetic."""
    sv = u.quantity.StaticValue(np.array([1.0, 2.0]))
    sv2 = u.quantity.StaticValue(np.array([3.0, 4.0]))
    assert np.allclose(np.asarray(sv + 1.0), np.array([2.0, 3.0]))
    assert np.allclose(np.asarray(1.0 + sv), np.array([2.0, 3.0]))
    assert np.allclose(np.asarray(sv - 1.0), np.array([0.0, 1.0]))
    assert np.allclose(np.asarray(3.0 - sv), np.array([2.0, 1.0]))
    assert np.allclose(np.asarray(sv * 2.0), np.array([2.0, 4.0]))
    assert np.allclose(np.asarray(2.0 * sv), np.array([2.0, 4.0]))
    assert np.allclose(np.asarray(sv / 2.0), np.array([0.5, 1.0]))
    assert np.allclose(np.asarray(4.0 / sv), np.array([4.0, 2.0]))
    assert np.allclose(np.asarray(sv % 2.0), np.array([1.0, 0.0]))

    sum_sv = sv + sv2
    assert isinstance(sum_sv, u.quantity.StaticValue)
    assert np.allclose(np.asarray(sum_sv.array), np.array([4.0, 6.0]))


def test_static_value_protocols() -> None:
    """StaticValue exposes NumPy-style protocols."""
    arr = np.array([1, 2], dtype=np.int32)
    sv = u.quantity.StaticValue(arr)

    assert np.array_equal(np.asarray(sv), arr)
    assert np.asarray(sv, dtype=np.float64).dtype == np.float64
    assert len(sv) == 2
    assert list(iter(sv)) == [1, 2]
    assert sv[0] == 1
    assert sv.shape == (2,)
    assert sv.dtype == jnp.asarray(arr).dtype
    assert repr(sv) == f"StaticValue({repr(arr)!s})"
    assert np.allclose(np.asarray(sv.sum()), np.sum(arr))


def test_static_value_eq_hash() -> None:
    """StaticValue equality and hashing follow the wrapped array."""
    sv1 = u.quantity.StaticValue(np.array([1.0, 2.0]))
    sv2 = u.quantity.StaticValue(np.array([1.0, 2.0]))
    sv3 = u.quantity.StaticValue(np.array([1.0, 3.0]))

    assert sv1 == sv2
    assert sv1 != sv3
    assert sv1 != 1
    assert sv1.__eq__(1) is NotImplemented
    assert hash(sv1) == hash(sv2)


def test_static_value_rich_comparisons() -> None:
    """StaticValue supports all rich comparison operators."""
    sv1 = u.quantity.StaticValue(np.array([1.0, 2.0, 3.0]))
    sv2 = u.quantity.StaticValue(np.array([2.0, 2.0, 2.0]))
    sv3 = u.quantity.StaticValue(np.array([1.0, 2.0, 3.0]))

    # Test with another StaticValue - eq/ne return bool
    assert (sv1 == sv3) is True
    assert (sv1 != sv2) is True
    assert (sv1 == sv2) is False

    # Ordering comparisons return arrays
    assert np.array_equal(sv1 < sv2, np.array([True, False, False]))
    assert np.array_equal(sv1 <= sv2, np.array([True, True, False]))
    assert np.array_equal(sv1 > sv2, np.array([False, False, True]))
    assert np.array_equal(sv1 >= sv2, np.array([False, True, True]))

    # Test with regular arrays
    arr = np.array([2.0, 2.0, 2.0])
    assert np.array_equal(sv1 == arr, np.array([False, True, False]))
    assert np.array_equal(sv1 != arr, np.array([True, False, True]))
    assert np.array_equal(sv1 < arr, np.array([True, False, False]))
    assert np.array_equal(sv1 <= arr, np.array([True, True, False]))
    assert np.array_equal(sv1 > arr, np.array([False, False, True]))
    assert np.array_equal(sv1 >= arr, np.array([False, True, True]))

    # Test with scalars
    assert np.array_equal(sv1 < 2.5, np.array([True, True, False]))
    assert np.array_equal(sv1 > 1.5, np.array([False, True, True]))


def test_static_value_comparison_incompatible_types() -> None:
    """StaticValue handles incompatible types correctly in comparisons."""
    sv = u.quantity.StaticValue(np.array([1.0, 2.0]))

    # __eq__ and __ne__ with incompatible types return NotImplemented
    assert sv.__eq__(object()) is NotImplemented
    assert sv.__eq__(1) is NotImplemented
    assert sv.__eq__("string") is NotImplemented
    assert sv.__eq__({"key": "value"}) is NotImplemented

    assert sv.__ne__(object()) is NotImplemented
    assert sv.__ne__(1) is NotImplemented
    assert sv.__ne__("string") is NotImplemented

    # Python's comparison fallback handles NotImplemented correctly
    # When comparing sv == 1, Python tries sv.__eq__(1), gets NotImplemented,
    # then tries (1).__eq__(sv), which returns False
    assert (sv == 1) is False
    assert (sv != 1) is True
    assert (sv == object()) is False
    assert (sv != object()) is True

    # Ordering comparisons with scalars that work with NumPy arrays
    # These delegate to NumPy's broadcasting rules
    assert np.array_equal(sv < 1.5, np.array([True, False]))
    assert np.array_equal(sv <= 2.0, np.array([True, True]))
    assert np.array_equal(sv > 1.5, np.array([False, True]))
    assert np.array_equal(sv >= 1.0, np.array([True, True]))

    # Ordering comparisons with incompatible types may raise errors
    # (this is expected NumPy behavior)
    with pytest.raises((TypeError, np.exceptions.DTypePromotionError)):
        sv < "string"  # noqa: B015

    with pytest.raises((TypeError, np.exceptions.DTypePromotionError)):
        sv <= object()  # noqa: B015


def test_static_value_from_dispatch() -> None:
    """StaticValue.from_ dispatches to expected constructors."""
    sv = u.quantity.StaticValue.from_([1.0, 2.0])
    assert isinstance(sv, u.quantity.StaticValue)
    assert np.array_equal(np.asarray(sv.array), np.array([1.0, 2.0]))

    sv2 = u.quantity.StaticValue.from_(sv)
    assert sv2 is sv

    with pytest.raises(TypeError, match="StaticQuantity does not accept JAX arrays"):
        u.quantity.StaticValue.from_(jnp.array([1.0, 2.0]))


def test_static_value_unary_ops() -> None:
    """StaticValue unary ops degrade to array values."""
    sv = u.quantity.StaticValue(np.array([-1.0, 2.0]))
    assert np.allclose(np.asarray(-sv), np.array([1.0, -2.0]))
    assert np.allclose(np.asarray(+sv), np.array([-1.0, 2.0]))
    assert np.allclose(np.asarray(abs(sv)), np.array([1.0, 2.0]))


def test_static_value_more_binary_ops() -> None:
    """StaticValue supports additional binary operators."""
    sv = u.quantity.StaticValue(np.array([3.0, 4.0]))
    assert np.allclose(np.asarray(sv // 2), np.array([1.0, 2.0]))
    assert np.allclose(np.asarray(5.0 // sv), np.array([1.0, 1.0]))
    assert np.allclose(np.asarray(sv**2), np.array([9.0, 16.0]))
    assert np.allclose(np.asarray(2.0**sv), np.array([8.0, 16.0]))
    assert np.allclose(np.asarray(5.0 % sv), np.array([2.0, 1.0]))

    sv_mat = u.quantity.StaticValue(np.array([[1.0, 2.0], [3.0, 4.0]]))
    sv_vec = u.quantity.StaticValue(np.array([1.0, 1.0]))
    out = sv_mat @ sv_vec
    assert isinstance(out, u.quantity.StaticValue)
    assert np.allclose(np.asarray(out.array), np.array([3.0, 7.0]))

    out_r = jnp.array([1.0, 1.0]) @ sv_mat
    assert np.allclose(np.asarray(out_r), np.array([4.0, 6.0]))

    rsum = sv.__radd__(u.quantity.StaticValue(np.array([1.0, 1.0])))
    assert isinstance(rsum, u.quantity.StaticValue)
    assert np.allclose(np.asarray(rsum.array), np.array([4.0, 5.0]))


def test_static_value_dunder_ops_explicit() -> None:
    """Exercise StaticValue dunder operators directly."""
    sv = u.quantity.StaticValue(np.array([3.0, 4.0]))
    assert np.allclose(np.asarray(sv.__floordiv__(2)), np.array([1.0, 2.0]))
    assert np.allclose(np.asarray(sv.__rfloordiv__(5.0)), np.array([1.0, 1.0]))
    assert np.allclose(np.asarray(sv.__pow__(2)), np.array([9.0, 16.0]))
    assert np.allclose(np.asarray(sv.__rpow__(2.0)), np.array([8.0, 16.0]))
    assert np.allclose(np.asarray(sv.__mod__(2.0)), np.array([1.0, 0.0]))
    assert np.allclose(np.asarray(sv.__rmod__(5.0)), np.array([2.0, 1.0]))

    sv_mat = u.quantity.StaticValue(np.array([[1.0, 2.0], [3.0, 4.0]]))
    sv_vec = u.quantity.StaticValue(np.array([1.0, 1.0]))
    out = sv_mat.__matmul__(sv_vec)
    assert isinstance(out, u.quantity.StaticValue)
    assert np.allclose(np.asarray(out.array), np.array([3.0, 7.0]))

    out_r = sv_mat.__rmatmul__(np.array([1.0, 1.0]))
    assert np.allclose(np.asarray(out_r), np.array([4.0, 6.0]))

    assert np.allclose(np.asarray(sv.__neg__()), np.array([-3.0, -4.0]))
    assert np.allclose(np.asarray(sv.__pos__()), np.array([3.0, 4.0]))
    assert np.allclose(np.asarray(sv.__abs__()), np.array([3.0, 4.0]))


@given(
    arr=np_arrays(
        dtype=np.float32,
        shape=st.integers(min_value=1, max_value=5),
        elements=st.floats(
            min_value=-10.0,
            max_value=10.0,
            allow_nan=False,
            allow_infinity=False,
            allow_subnormal=False,
            width=32,
        ),
    )
)
def test_static_value_ops_property(arr: np.ndarray) -> None:
    """StaticValue ops match array results for sampled inputs."""
    sv = u.quantity.StaticValue(arr)
    sv2 = u.quantity.StaticValue(arr)

    out = sv + sv2
    assert isinstance(out, u.quantity.StaticValue)
    assert np.allclose(np.asarray(out.array), arr + arr)
    assert np.allclose(np.asarray(sv + 1.0), arr + 1.0)
    assert np.allclose(np.asarray(1.0 + sv), arr + 1.0)


def test_quantity_ops_with_static_value() -> None:
    """Quantity should operate correctly with StaticValue-backed values."""
    sv = u.quantity.StaticValue(np.array([1.0, 2.0]))
    q_static = u.Q(sv, "m")
    q = u.Q(jnp.array([3.0, 4.0]), "m")

    out_add = q_static + q
    assert np.allclose(np.asarray(out_add.value), np.array([4.0, 6.0]))

    out_mul = q_static * q
    assert np.allclose(np.asarray(out_mul.value), np.array([3.0, 8.0]))
    assert out_mul.unit == u.unit("m2")


def test_quantity_static_value_as_jit_static_arg() -> None:
    """Quantity(StaticValue) can be used as a static argument to jitted functions."""
    # Create a Quantity with StaticValue
    sv = u.quantity.StaticValue(np.array([2.0, 3.0]))
    q_static = u.Q(sv, "m")

    # Define a function that uses the static quantity
    @partial(jax.jit, static_argnames=("q_static",))
    def compute(x, q_static):
        # Access the value from the static quantity
        return x * jnp.asarray(q_static.value)

    # First call - should compile
    x = jnp.array([1.0, 1.0])
    result = compute(x, q_static)
    assert np.allclose(result, np.array([2.0, 3.0]))

    # Second call with same static value - should use cached compilation
    result2 = compute(jnp.array([2.0, 2.0]), q_static)
    assert np.allclose(result2, np.array([4.0, 6.0]))

    # Third call with different static value - should trigger recompilation
    sv_new = u.quantity.StaticValue(np.array([5.0, 7.0]))
    q_static_new = u.Q(sv_new, "m")
    result3 = compute(jnp.array([1.0, 1.0]), q_static_new)
    assert np.allclose(result3, np.array([5.0, 7.0]))

    # Verify the quantity is actually being used as static (must be hashable)
    assert isinstance(hash(q_static), int)
    assert isinstance(hash(q_static_new), int)


def test_quantity_static_value_jit_with_operations() -> None:
    """Quantity(StaticValue) works correctly in jitted operations."""
    sv = u.quantity.StaticValue(np.array([2.0, 3.0]))
    q_static = u.Q(sv, "m")

    @partial(jax.jit, static_argnames=("scale",))
    def scale_and_add(x, scale):
        # Convert scale to array for computation
        scale_array = jnp.asarray(scale.value)
        return x + scale_array

    x = jnp.array([1.0, 2.0])
    result = scale_and_add(x, q_static)
    assert np.allclose(result, np.array([3.0, 5.0]))

    # Test with different units to ensure unit is also part of static args
    sv2 = u.quantity.StaticValue(np.array([2.0, 3.0]))
    q_static2 = u.Q(sv2, "km")

    result2 = scale_and_add(x, q_static2)
    assert np.allclose(result2, np.array([3.0, 5.0]))

    # Verify they have different hashes (different units)
    assert hash(q_static) != hash(q_static2)



"""Tests for Quantity printing with wadler-lindig."""

import wadler_lindig as wl

import unxt as u


class TestShortName:
    """Test the short_name feature for wadler-lindig printing."""

    def test_quantity_has_short_name(self):
        """Test that Quantity has a short_name class variable."""
        assert hasattr(u.Quantity, "short_name")
        assert u.Quantity.short_name == "Q"

    def test_barequantity_no_short_name(self):
        """Test that BareQuantity doesn't have a short_name or it's None."""
        # It should either not have the attribute or have it as None
        short_name = getattr(u.quantity.BareQuantity, "short_name", None)
        assert short_name is None

    def test_use_short_name_default_false(self):
        """Test that use_short_name defaults to False."""
        q = u.Q([1, 2, 3], "m")
        result = wl.pformat(q)
        assert result.startswith("Quantity")
        assert not result.startswith("Q(")

    def test_use_short_name_true(self):
        """Test that use_short_name=True uses the short name."""
        q = u.Q([1, 2, 3], "m")
        result = wl.pformat(q, use_short_name=True)
        assert result.startswith("Q(")
        assert "unit='m'" in result

    def test_use_short_name_with_include_params(self):
        """Test that use_short_name works with include_params."""
        q = u.Q([1, 2, 3], "m")
        result = wl.pformat(q, use_short_name=True, include_params=True)
        assert result.startswith("Q['length']")

    def test_use_short_name_with_named_unit_false(self):
        """Test that use_short_name works with named_unit=False."""
        q = u.Q([1, 2, 3], "m")
        result = wl.pformat(q, use_short_name=True, named_unit=False)
        assert result.startswith("Q(")
        # Should have unit as positional arg not named
        assert "'m')" in result or ", 'm')" in result

    def test_use_short_name_with_short_arrays(self):
        """Test that use_short_name works with short_arrays."""
        q = u.Q([1, 2, 3], "m")

        # Default short_arrays=True
        result = wl.pformat(q, use_short_name=True, short_arrays=True)
        assert result.startswith("Q(")
        assert "i32[3]" in result

        # short_arrays=False
        result = wl.pformat(q, use_short_name=True, short_arrays=False)
        assert result.startswith("Q(")
        assert "Array(" in result

    def test_use_short_name_with_short_arrays_compact(self):
        """Test that use_short_name works with short_arrays='compact'."""
        q = u.Q([1, 2, 3], "m")
        result = wl.pformat(q, use_short_name=True, short_arrays="compact")
        assert result.startswith("Q(")
        assert "[1, 2, 3]" in result

    def test_bare_quantity_use_short_name_none(self):
        """Test that BareQuantity with use_short_name=True still uses full name."""
        q = u.quantity.BareQuantity([1, 2, 3], "m")
        result = wl.pformat(q, use_short_name=True)
        # Should still use full name since short_name is None
        assert result.startswith("BareQuantity")

    def test_pprint(self):
        """Test that pprint works with use_short_name."""
        q = u.Q([1, 2, 3], "m")
        # This should not raise an error
        wl.pprint(q, use_short_name=True)

    def test_pdoc_method_directly(self):
        """Test calling __pdoc__ directly with use_short_name."""
        q = u.Q([1, 2, 3], "m")

        doc = q.__pdoc__(use_short_name=False)
        formatted = wl.pformat(doc)
        assert formatted.startswith("Quantity")

        doc = q.__pdoc__(use_short_name=True)
        formatted = wl.pformat(doc)



"""Working with units.

The main features are:

- ``unxt.units.unit``: a function to construct units.
- ``unxt.units.unit_of``: a function to get the units of an object.

"""

__all__ = ("unit", "unit_of", "AbstractUnit")

from .setup_package import install_import_hook

with install_import_hook("unxt.units"):
    from ._src.units import AbstractUnit
    from unxt_api import unit, unit_of

# Clean up the namespace
del install_import_hook



"""Quantities in JAX.

This module provides JAX-compatible quantity types with automatic unit handling,
conversions, and a rich set of mathematical operations that preserve dimensional
correctness.

## Core Classes

- **`Quantity`**: The main quantity class with dimension parametrization and
  full unit checking. Aliased as `Q` for convenience.
- **`StaticQuantity`**: Parametric quantity with static NumPy values for JAX
  static arguments.
- **`BareQuantity`**: Lightweight quantity without dimension parametrization.
- **`Angle`**: Specialized quantity type for angular measurements with wrapping
  support.

## Key Functions

- **`uconvert`**: Convert a quantity to different units.
- **`ustrip`**: Extract the numerical value from a quantity.
- **`is_unit_convertible`**: Check if two units are convertible to each other.
- **`wrap_to`**: Wrap angular quantities to a specified range.

## Examples

### Creating quantities with units:

>>> import unxt as u
>>> import jax
>>> import jax.numpy as jnp

>>> distance = u.Quantity(100, "m")
>>> distance
Quantity(Array(100, dtype=int32, ...), unit='m')

Create from arrays

>>> velocities = u.Quantity([10, 20, 30], "m/s")
>>> velocities
Quantity(Array([10, 20, 30], dtype=int32), unit='m / s')

Unit conversions and arithmetic:

Convert units

>>> distance_km = u.uconvert("km", distance)
>>> distance_km
Quantity(Array(0.1, dtype=float32, ...), unit='km')

Arithmetic preserves units.

>>> time = u.Q(5, "s")  # use Quantity alias
>>> velocity = distance / time
>>> velocity
Quantity(Array(20., dtype=float32, ...), unit='m / s')

Strip units for numerical operations

>>> u.ustrip("m", distance)
Array(100, dtype=int32, ...)


### Working with angles:

Create angle quantities

>>> theta = u.Angle(180, "deg")
>>> theta
Angle(Array(180, dtype=int32, ...), unit='deg')

Convert to radians

>>> theta_rad = u.uconvert("rad", theta)
>>> theta_rad
Angle(Array(3.1415927, dtype=float32, ...), unit='rad')

Wrap angles to a range

>>> angle = u.Angle(450, "deg")
>>> wrapped = u.quantity.wrap_to(angle, u.Angle(0, "deg"), u.Angle(360, "deg"))
>>> wrapped
Angle(Array(90, dtype=int32, ...), unit='deg')


### Advanced usage with JAX transformations:

Quantities work with JAX transformations

>>> def kinetic_energy(mass, velocity):
...     return 0.5 * mass * velocity**2

>>> mass = u.Q(2.0, "kg")
>>> vel = u.Q(10.0, "m/s")
>>> energy = kinetic_energy(mass, vel)
>>> energy
Quantity(Array(100., dtype=float32, ...), unit='m2 kg / s2')

Convert to standard energy units

>>> u.uconvert("J", energy)
Quantity(Array(100., dtype=float32, ...), unit='m2 kg / s2')

JIT compilation works seamlessly

>>> @jax.jit
... def compute_force(mass, accel):
...     return mass * accel
>>> force = compute_force(u.Q(5.0, "kg"), u.Q(9.8, "m/s^2"))
>>> force
Quantity(Array(49., dtype=float32, ...), unit='kg m / s2')

"""
# pylint: disable=duplicate-code

__all__ = (
    # Core
    "Quantity",
    "Q",  # convenience alias
    "StaticQuantity",
    "StaticValue",
    # Base
    "AbstractQuantity",
    # Fast
    "BareQuantity",
    # Angles
    "AbstractAngle",
    "Angle",
    # Base Parametric
    "AbstractParametricQuantity",
    # Functional
    "uconvert_value",
    "uconvert",
    "ustrip",
    "is_unit_convertible",
    "wrap_to",
    "is_any_quantity",
    "convert_to_quantity_value",
    "AllowValue",
)


from .setup_package import install_import_hook

with install_import_hook("unxt.quantity"):
    from ._src.quantity.angle import Angle
    from ._src.quantity.base import AbstractQuantity, is_any_quantity
    from ._src.quantity.base_angle import AbstractAngle
    from ._src.quantity.base_parametric import AbstractParametricQuantity
    from ._src.quantity.flag import AllowValue
    from ._src.quantity.quantity import Q, Quantity
    from ._src.quantity.static_quantity import StaticQuantity
    from ._src.quantity.unchecked import BareQuantity
    from ._src.quantity.value import StaticValue, convert_to_quantity_value
    from unxt_api import is_unit_convertible, uconvert, uconvert_value, ustrip, wrap_to

    # isort: split
    # Register dispatches and conversions
    from ._src.quantity import (
        register_api,
        register_conversions,
        register_dispatches,
        register_primitives,
    )

# Clean up namespace
del (
    register_conversions,
    register_api,
    register_dispatches,
    register_primitives,
    install_import_hook,
)



"""Dimension API for unxt.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__ = ("dimension", "dimension_of")

from typing import Any

import plum


@plum.dispatch.abstract
def dimension(obj: Any, /) -> Any:
    """Construct the dimension.

    .. note::

        This function uses multiple dispatch. Dispatches made in other modules
        may not be included in the rendered docs. To see the full range of
        options, execute ``unxt.dims.dimension.methods`` in an interactive
        Python session.

    """


@plum.dispatch.abstract
def dimension_of(obj: Any, /) -> Any:
    """Return the dimension of the given units.

    .. note::

        This function uses multiple dispatch. Dispatches made in other modules
        may not be included in the rendered docs. To see the full range of
        options, execute ``unxt.dimension_of.methods`` in an interactive Python
        session.

    """



"""Package setup information.

Note that this module is NOT public API nor are any of its contents.
Stability is NOT guaranteed.
This module exposes package setup information for the `unxt` package.

"""

__all__: tuple[str, ...] = ()

import contextlib
import os
from collections.abc import Sequence
from typing import Any, Final, Literal

from jaxtyping import install_import_hook as _install_import_hook

_RUNTIME_TYPECHECKER: str | None | Literal[False]
match os.getenv("UNXT_ENABLE_RUNTIME_TYPECHECKING", "False"):
    case "False":
        _RUNTIME_TYPECHECKER = False
    case "None":
        _RUNTIME_TYPECHECKER = None
    case str() as _name:
        _RUNTIME_TYPECHECKER = _name

RUNTIME_TYPECHECKER: Final[str | None | Literal[False]] = _RUNTIME_TYPECHECKER
"""Runtime type checking variable "UNXT_ENABLE_RUNTIME_TYPECHECKING".

Set to "False" to disable runtime typechecking (default).
Set to "None" to only enable typechecking for `@jaxtyped`-decorated functions.
Set to "beartype.beartype" to enable runtime typechecking.

See https://docs.kidger.site/jaxtyping/api/runtime-type-checking for more
information on options.

"""


def install_import_hook(
    modules: str | Sequence[str], /
) -> contextlib.AbstractContextManager[Any, None]:
    """Install the jaxtyping import hook for the given modules.

    Parameters
    ----------
    modules
        Module name or sequence of module names to install the import hook for.

    Returns
    -------
    contextlib.AbstractContextManager
        Context manager that installs the import hook on entry and removes it on exit.

    """
    return (
        _install_import_hook(modules, RUNTIME_TYPECHECKER)
        if RUNTIME_TYPECHECKER is not False
        else contextlib.nullcontext()
    )



"""Astropy units compatibility."""

__all__: tuple[str, ...] = ()


import dataclasses

import astropy.units as apyu
import plum

from .custom_types import APYUnits

# ===================================================================
# Register dispatches


@plum.dispatch
def unit(obj: apyu.UnitBase | apyu.Unit, /) -> APYUnits:
    """Construct the units from an Astropy unit.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u
    >>> u.unit(apyu.km)
    Unit("km")

    """
    return apyu.Unit(obj)


@plum.dispatch
def unit(obj: apyu.Quantity, /) -> APYUnits:
    """Construct the units from an Astropy quantity.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u
    >>> u.unit(apyu.Quantity(2, "km"))
    Unit("2 km")

    """
    return apyu.Unit(obj)


# -------------------------------------------------------------------


@plum.dispatch
def unit_of(obj: apyu.UnitBase | apyu.Unit, /) -> APYUnits:
    """Return the units of an object.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> u.unit_of(apyu.km)
    Unit("km")

    """
    return obj


@plum.dispatch
def unit_of(obj: apyu.Quantity, /) -> APYUnits:
    """Return the units of an Astropy quantity.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> u.unit_of(apyu.Quantity(1, "km"))
    Unit("km")

    """
    return unit_of(obj.unit)


# ===================================================
# `Dataclassish` support


@plum.dispatch
def fields(obj: APYUnits, /) -> tuple[dataclasses.Field, ...]:
    """Return the fields of a dimension.

    Examples
    --------
    >>> import dataclassish as dc
    >>> import astropy.units as apyu

    >>> dim = apyu.Unit("m")
    >>> dc.fields(dim)
    (Field(name='_names',...)

    """
    st_field = dataclasses.Field(
        dataclasses.MISSING,
        dataclasses.MISSING,
        True,  # noqa: FBT003
        True,  # noqa: FBT003
        True,  # noqa: FBT003
        True,  # noqa: FBT003
        {},
        False,  # noqa: FBT003
    )
    st_field.name = "_names"
    return (st_field,)



# pylint: disable=import-error, no-member, unsubscriptable-object
#    b/c it doesn't understand dataclass fields

__all__ = ("Quantity", "Q")

from typing import ClassVar, final

import equinox as eqx
from jaxtyping import Array, Shaped
from plum import parametric

from .base_parametric import AbstractParametricQuantity
from .value import StaticValue, convert_to_quantity_value
from unxt.units import AbstractUnit, unit as parse_unit


@final
@parametric
class Quantity(AbstractParametricQuantity):
    """Arrays with associated units.

    This class is parametrized by the dimensions of the units.

    Attributes
    ----------
    short_name : str
        Short name 'Q' used for compact wadler-lindig printing.

    Examples
    --------
    >>> import unxt as u

    From an integer:

    >>> u.Quantity(1, "m")
    Quantity(Array(1, dtype=int32, ...), unit='m')

    From a float:

    >>> u.Q(1.0, "m")
    Quantity(Array(1., dtype=float32, ...), unit='m')

    From a list:

    >>> u.Quantity([1, 2, 3], "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    From a tuple:

    >>> u.Quantity((1, 2, 3), "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    From a `numpy.ndarray`:

    >>> import numpy as np
    >>> u.Quantity(np.array([1, 2, 3]), "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    From a `jax.Array`:

    >>> import jax.numpy as jnp
    >>> u.Quantity(jnp.array([1, 2, 3]), "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    The unit can also be given as a units object:

    >>> u.Quantity(1, u.unit("m"))
    Quantity(Array(1, dtype=int32, ...), unit='m')

    In the previous examples, the dimension parameter was inferred from the
    values. It can also be given explicitly:

    >>> u.Quantity["length"](1, "m")
    Quantity(Array(1, dtype=int32, ...), unit='m')

    This can be used for runtime checking of the input dimension!

    >>> try:
    ...     u.Quantity["length"](1, "s")
    ... except Exception as e:
    ...     print(e)
    Physical type mismatch.

    The dimension can also be given as a dimension object:

    >>> dims = u.dimension("length")
    >>> dims
    PhysicalType('length')
    >>> u.Quantity[dims](1.0, "m")
    Quantity(Array(1., dtype=float32, ...), unit='m')

    Or as a unit:

    >>> u.Quantity[u.unit("m")](1.0, "m")
    Quantity(Array(1., dtype=float32, ...), unit='m')

    Some tricky cases are when the physical type is unknown:

    >>> unit = u.unit("m2 / (kg s2)")
    >>> u.dimension_of(unit)
    PhysicalType('unknown')

    The dimension can be given as a string in all cases, but is necessary when
    the physical type is unknown:

    >>> print(u.Quantity["m2 kg-1 s-2"](1.0, unit))  # to show the [dim]
    Quantity['m2 kg-1 s-2'](1., unit='m2 / (kg s2)')

    """

    short_name: ClassVar[str] = "Q"
    """Short name for compact printing."""

    value: Shaped[Array, "*shape"] | StaticValue = eqx.field(
        converter=convert_to_quantity_value
    )
    """The value of the `AbstractQuantity`."""

    unit: AbstractUnit = eqx.field(static=True, converter=parse_unit)
    """The unit associated with this value."""


Q = Quantity  # convenience alias



"""Benchmark tests for `unxt.uconvert_value` with JAX transformations.

These benchmarks test uconvert_value performance with jit and vmap
using pytest-codspeed for precise measurements.
"""

import jax
import jax.numpy as jnp
import pytest

import unxt as u


@pytest.fixture
def xarray():
    return jnp.linspace(1.0, 1000.0, num=int(1e4), dtype=jnp.float32)


#####################################################################
# Benchmark: uconvert_value with scalar values


@pytest.mark.benchmark(group="uconvert_value", max_time=1.0)
def test_uconvert_value_scalar_string_units(benchmark):
    """Benchmark uconvert_value with scalar value and string units."""
    benchmark(u.uconvert_value, "m", "km", 5.0)


@pytest.mark.benchmark(group="uconvert_value", max_time=1.0)
def test_uconvert_value_scalar_unit_objects(benchmark):
    """Benchmark uconvert_value with scalar value and unit objects."""
    uto = u.unit("m")
    ufrom = u.unit("km")
    benchmark(u.uconvert_value, uto, ufrom, 5.0)


#####################################################################
# Benchmark: uconvert_value with arrays (no JIT)


@pytest.mark.benchmark(group="uconvert_value", max_time=1.0)
def test_uconvert_value_array_no_jit(benchmark, xarray):
    """Benchmark uconvert_value with array (no JIT compilation)."""
    benchmark(u.uconvert_value, u.unit("m"), u.unit("km"), xarray)


#####################################################################
# Benchmark: uconvert_value with vmap only


@pytest.fixture
def func_vmap_uconvert():
    """Create vmap version of uconvert_value."""
    return jax.vmap(u.uconvert_value, in_axes=(None, None, 0))


@pytest.mark.benchmark(group="uconvert_value_vmap", max_time=1.0)
def test_uconvert_value_vmap_no_jit(benchmark, func_vmap_uconvert, xarray):
    """Benchmark vmap(uconvert_value) without JIT."""
    benchmark(func_vmap_uconvert, u.unit("m"), u.unit("km"), xarray)


@pytest.mark.benchmark(group="uconvert_value_vmap", max_time=1.0)
def test_uconvert_value_vmap_jit_compile(xarray):
    """Benchmark vmap(uconvert_value) JIT compilation."""
    vmap_jit_uconvert = jax.vmap(
        jax.jit(u.uconvert_value, static_argnums=(0, 1)),
        in_axes=(None, None, 0),
    )
    _ = vmap_jit_uconvert.__wrapped__.lower(u.unit("m"), u.unit("km"), xarray).compile()


@pytest.mark.benchmark(group="uconvert_value_vmap", max_time=1.0)
def test_uconvert_value_vmap_jit_execute(benchmark, func_vmap_uconvert, xarray):
    """Benchmark vmap(uconvert_value) after JIT compilation."""
    jit_fn = jax.jit(func_vmap_uconvert)
    # Warm up
    _ = jit_fn(u.unit("m"), u.unit("km"), xarray)
    # Benchmark
    benchmark(lambda: jax.block_until_ready(jit_fn(u.unit("m"), u.unit("km"), xarray)))


#####################################################################
# Benchmark: uconvert_value with vmap + jit (static args)


@pytest.fixture
def func_vmap_jit_uconvert():
    """Create vmap(jit(uconvert_value)) with static args."""
    return jax.vmap(
        jax.jit(u.uconvert_value, static_argnums=(0, 1)),
        in_axes=(None, None, 0),
    )


@pytest.mark.benchmark(group="uconvert_value_vmap_jit", max_time=1.0)
def test_uconvert_value_vmap_jit_execute(benchmark, func_vmap_jit_uconvert, xarray):
    """Benchmark vmap(jit(uconvert_value)) with static unit arguments."""
    uto = u.unit("m")
    ufrom = u.unit("km")
    # Warm up
    _ = func_vmap_jit_uconvert(uto, ufrom, xarray)
    # Benchmark
    benchmark(lambda: jax.block_until_ready(func_vmap_jit_uconvert(uto, ufrom, xarray)))


#####################################################################
# Benchmark: Larger array performance


@pytest.mark.benchmark(group="uconvert_value_large", max_time=1.0)
def test_uconvert_value_large_array_no_jit(benchmark, xarray):
    """Benchmark uconvert_value with large array (no JIT)."""
    benchmark(u.uconvert_value, u.unit("m"), u.unit("km"), xarray)


@pytest.mark.benchmark(group="uconvert_value_large", max_time=1.0)
def test_uconvert_value_large_array_vmap_jit(benchmark, func_vmap_jit_uconvert, xarray):
    """Benchmark vmap(jit(uconvert_value)) with large array."""
    uto = u.unit("m")
    ufrom = u.unit("km")
    # Warm up
    _ = func_vmap_jit_uconvert(uto, ufrom, xarray)
    # Benchmark
    benchmark(lambda: jax.block_until_ready(func_vmap_jit_uconvert(uto, ufrom, xarray)))



"""Abstract dispatch API for unxt.

This package defines the abstract dispatch interfaces for unxt's core functionality.
"""

__all__ = (
    "__version__",
    "dimension",
    "dimension_of",
    "uconvert_value",
    "uconvert",
    "ustrip",
    "is_unit_convertible",
    "wrap_to",
    "unit",
    "unit_of",
    "unitsystem_of",
)

from ._src.dimensions import dimension, dimension_of
from ._src.quantity import (
    is_unit_convertible,
    uconvert,
    uconvert_value,
    ustrip,
    wrap_to,
)
from ._src.units import unit, unit_of
from ._src.unitsystems import unitsystem_of
from ._version import version as __version__



"""Integration tests for uconvert_value JAX transformations and performance.

This module tests uconvert_value with JAX transformations (jit, vmap)
and demonstrates the performance improvement from JIT compilation.
"""

import jax
import jax.numpy as jnp
import numpy as np

import unxt as u


class TestUconvertValueVmapJit:
    """Test uconvert_value with vmap and jit transformations."""

    def test_uconvert_value_vmap_jit_basic(self) -> None:
        """Test basic vmap(jit(uconvert_value)) functionality.

        With vmap(jit(uconvert_value)), we can convert multiple values
        efficiently by:
        - JIT compiling the function once
        - Using vmap to apply it over multiple input values
        - Keeping unit arguments static (not vmapped)
        """
        # Define vmapped and jitted version
        # vmap maps over x (last argument), leaving uto and ufrom unmapped
        # Use static_argnums=(0, 1) since parameters are positional-only
        vmap_jit_uconvert = jax.vmap(
            jax.jit(u.uconvert_value, static_argnums=(0, 1)),
            in_axes=(None, None, 0),  # (uto, ufrom, x)
        )

        # Input values to convert
        values = jnp.array([1.0, 2.0, 5.0, 10.0])

        # Convert km to m
        result = vmap_jit_uconvert(u.unit("m"), u.unit("km"), values)

        # Check results
        expected = jnp.array([1000.0, 2000.0, 5000.0, 10000.0])
        assert np.allclose(result, expected)

    def test_uconvert_value_vmap_jit_with_string_units(self) -> None:
        """Test vmap(jit(uconvert_value)) with string units."""
        vmap_jit_uconvert = jax.vmap(
            jax.jit(u.uconvert_value, static_argnums=(0, 1)),
            in_axes=(None, None, 0),
        )

        values = jnp.array([1000.0, 2000.0, 3000.0])
        result = vmap_jit_uconvert("km", "m", values)

        expected = jnp.array([1.0, 2.0, 3.0])
        assert np.allclose(result, expected)


class TestUconvertValueJitOnly:
    """Test uconvert_value with JIT alone (without vmap)."""

    def test_uconvert_value_jit_single_value(self) -> None:
        """Test JIT compilation of uconvert_value for single value."""
        jit_uconvert = jax.jit(u.uconvert_value, static_argnums=(0, 1))

        result = jit_uconvert(u.unit("m"), u.unit("km"), 5.0)
        assert jnp.isclose(result, 5000.0)

    def test_uconvert_value_jit_with_array(self) -> None:
        """Test JIT compilation of uconvert_value for array values."""
        jit_uconvert = jax.jit(u.uconvert_value, static_argnums=(0, 1))

        values = jnp.array([1.0, 2.0, 3.0])
        result = jit_uconvert(u.unit("m"), u.unit("km"), values)

        expected = jnp.array([1000.0, 2000.0, 3000.0])
        assert np.allclose(result, expected)


class TestUconvertValueVmapOnly:
    """Test uconvert_value with vmap alone (without JIT)."""

    def test_uconvert_value_vmap_vectorization(self) -> None:
        """Test vmap vectorization of uconvert_value."""
        vmap_uconvert = jax.vmap(u.uconvert_value, in_axes=(None, None, 0))

        values = jnp.array([1.0, 2.0, 5.0])
        result = vmap_uconvert(u.unit("m"), u.unit("km"), values)

        expected = jnp.array([1000.0, 2000.0, 5000.0])
        assert np.allclose(result, expected)

    def test_uconvert_value_nested_vmap(self) -> None:
        """Test nested vmap over multiple dimensions."""
        # Nested vmap: outer maps over axis 0, inner maps over axis 0
        # For 2D input, this vmaps over rows, then elements within rows
        vmap_uconvert = jax.vmap(
            jax.vmap(u.uconvert_value, in_axes=(None, None, 0)),
            in_axes=(None, None, 0),
        )

        # 2D array: (batches, values_per_batch)
        values = jnp.array([[1.0, 2.0], [3.0, 4.0]])

        result = vmap_uconvert(u.unit("m"), u.unit("km"), values)

        expected = jnp.array([[1000.0, 2000.0], [3000.0, 4000.0]])
        assert np.allclose(result, expected)



"""Quantity compatibility."""

__all__: tuple[str, ...] = ()

import dataclasses
from typing import Any, NoReturn

import plum
from astropy.coordinates import Angle as AstropyAngle, Distance as AstropyDistance
from astropy.units import Quantity as AstropyQuantity
from jaxtyping import ArrayLike

import dataclassish as dc
import quaxed.numpy as jnp

import unxt_api as uapi
from .custom_types import APYUnits
from unxt.quantity import AbstractQuantity, AllowValue, BareQuantity, Quantity, ustrip

# ============================================================================
# Value Converter


@plum.dispatch
def convert_to_quantity_value(obj: AstropyQuantity, /) -> NoReturn:
    """Disallow conversion of `AstropyQuantity` to a value.

    >>> import astropy.units as apyu
    >>> from unxt.quantity import convert_to_quantity_value

    >>> try:
    ...     convert_to_quantity_value(apyu.Quantity(1, "m"))
    ... except TypeError as e:
    ...     print(e)
    Cannot convert 'Quantity' to a value.
    For a Quantity, use the `.from_` constructor instead.

    """
    msg = (
        f"Cannot convert {type(obj).__name__!r} to a value. "
        "For a Quantity, use the `.from_` constructor instead."
    )
    raise TypeError(msg)


# ============================================================================
# AbstractQuantity


# -----------------
# Constructor


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity], value: AstropyQuantity, /, **kwargs: Any
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> u.Quantity.from_(apyu.Quantity(1, "m"))
    Quantity(Array(1., dtype=float32), unit='m')

    """
    u = uapi.unit_of(value)
    value = jnp.asarray(uapi.ustrip(u, value), **kwargs)
    return cls(value, u)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity], value: AstropyQuantity, u: Any, /, **kwargs: Any
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> u.Quantity.from_(apyu.Quantity(1, "m"), "cm")
    Quantity(Array(100., dtype=float32), unit='cm')

    """
    u = uapi.unit(u)
    value = jnp.asarray(uapi.ustrip(u, value), **kwargs)
    return cls(value, u)


# -----------------
# Conversion Methods


@plum.conversion_method(type_from=AbstractQuantity, type_to=AstropyQuantity)  # type: ignore[arg-type]
def convert_unxt_quantity_to_astropy_quantity(
    q: AbstractQuantity, /
) -> AstropyQuantity:
    """Convert a `unxt.AbstractQuantity` to a `astropy.units.Quantity`.

    Examples
    --------
    >>> import unxt as u
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert

    >>> convert(u.Q(1.0, "cm"), AstropyQuantity)
    <Quantity 1. cm>

    """
    u = uapi.unit_of(q)
    return AstropyQuantity(uapi.ustrip(u, q), u)


@plum.conversion_method(type_from=AbstractQuantity, type_to=AstropyDistance)  # type: ignore[arg-type]
def convert_unxt_quantity_to_astropy_distance(
    q: AbstractQuantity, /
) -> AstropyDistance:
    """Convert a `unxt.AbstractQuantity` to a `astropy.coordinates.Distance`.

    Examples
    --------
    >>> import unxt as u
    >>> from astropy.coordinates import Distance as AstropyDistance
    >>> from plum import convert

    >>> convert(u.Q(1.0, "cm"), AstropyDistance)
    <Distance 1. cm>

    """
    u = uapi.unit_of(q)
    return AstropyDistance(uapi.ustrip(u, q), u)


@plum.conversion_method(type_from=AbstractQuantity, type_to=AstropyAngle)  # type: ignore[arg-type]
def convert_unxt_quantity_to_astropy_angle(q: AbstractQuantity, /) -> AstropyAngle:
    """Convert a `unxt.quantity.AbstractQuantity` to a `astropy.coordinates.Angle`.

    Examples
    --------
    >>> import unxt as u
    >>> from astropy.coordinates import Angle as AstropyAngle
    >>> from plum import convert

    >>> convert(u.Q(1.0, "rad"), AstropyAngle)
    <Angle 1. rad>

    """
    u = uapi.unit_of(q)
    return AstropyAngle(uapi.ustrip(u, q), u)


# ============================================================================
# Quantity


@plum.conversion_method(type_from=AstropyQuantity, type_to=Quantity)
def convert_astropy_quantity_to_unxt_quantity(q: AstropyQuantity, /) -> Quantity:
    """Convert a `astropy.units.Quantity` to a `unxt.Quantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt import Quantity

    >>> convert(AstropyQuantity(1.0, "cm"), Quantity)
    Quantity(Array(1., dtype=float32), unit='cm')

    """
    u = uapi.unit_of(q)
    return Quantity(uapi.ustrip(u, q), u)


# ============================================================================
# BareQuantity


@plum.conversion_method(type_from=AstropyQuantity, type_to=BareQuantity)  # type: ignore[arg-type]
def convert_astropy_quantity_to_unxt_barequantity(
    q: AstropyQuantity, /
) -> BareQuantity:
    """Convert a `astropy.units.Quantity` to a `unxt.BareQuantity`.

    Examples
    --------
    >>> from astropy.units import Quantity as AstropyQuantity
    >>> from plum import convert
    >>> from unxt.quantity import BareQuantity

    >>> convert(AstropyQuantity(1.0, "cm"), BareQuantity)
    BareQuantity(Array(1., dtype=float32), unit='cm')

    """
    u = uapi.unit_of(q)
    return BareQuantity(uapi.ustrip(u, q), u)


###############################################################################


@plum.dispatch
def uconvert_value(uto: APYUnits, ufrom: APYUnits, x: ArrayLike, /) -> ArrayLike:
    """Convert the value to the specified units.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> u.uconvert_value(apyu.Unit("km"), apyu.Unit("m"), 1000)
    1.0

    """
    # Hot-path: if no unit conversion is necessary
    if ufrom == uto:
        return x

    # Compute the new value.
    return ufrom.to(uto, x)


@plum.dispatch
def uconvert(u: APYUnits, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> x = u.Quantity(1000, "m")
    >>> u.uconvert(u.unit("km"), x)
    Quantity(Array(1., dtype=float32, ...), unit='km')

    >>> x = u.Quantity([1, 2, 3], "Kelvin")
    >>> with apyu.add_enabled_equivalencies(apyu.temperature()):
    ...     y = x.uconvert("deg_C")
    >>> y
    Quantity(Array([-272.15, -271.15, -270.15], dtype=float32, ...), unit='deg_C')

    >>> x = Quantity([1, 2, 3], "radian")
    >>> with apyu.add_enabled_equivalencies(apyu.dimensionless_angles()):
    ...     y = x.uconvert("")
    >>> y
    Quantity(Array([1., 2., 3.], dtype=float32, ...), unit='')

    """
    # Hot-path: if no unit conversion is necessary
    if x.unit == u:
        return x

    value = x.unit.to(u, uapi.ustrip(x))

    # If the dimensions are the same, we can just replace the value and unit.
    if uapi.dimension_of(x.unit) == uapi.dimension_of(u):
        return dc.replace(x, value=value, unit=u)

    # If the dimensions are different, we need to create a new quantity since
    # the dimensions can be part of the type. This won't work if the Quantity
    # type itself needs to be changed, e.g. `unxt.Angle` -> `unxt.Quantity`.
    # These cases are handled separately, in other dispatches.
    fs = dict(dc.field_items(x))  # pylint: disable=unreachable
    fs["value"] = value
    fs["unit"] = u

    return plum.type_unparametrized(x)(**fs)


# ============================================================================


@plum.dispatch
def ustrip(u: Any, x: AstropyQuantity) -> Any:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import astropy.units as apyu
    >>> import unxt as u

    >>> x = apyu.Quantity(1000, "m")
    >>> float(u.ustrip(u.unit("m"), x))
    1000.0

    """
    return x.to_value(u)


@plum.dispatch  # TODO: type annotate by value
def ustrip(flag: type[AllowValue], u: Any, x: AstropyQuantity, /) -> Any:
    """Strip the units from a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> q = u.Quantity(1000, "m")
    >>> u.ustrip(AllowValue, "km", q)
    Array(1., dtype=float32, ...)

    """
    return uapi.ustrip(u, x)


# ============================================================================


@plum.dispatch
def fields(obj: AbstractQuantity, /) -> tuple[dataclasses.Field, ...]:
    """Return the fields of a quantity.

    Examples
    --------
    >>> import dataclassish as dc
    >>> import unxt as u

    >>> q = u.Quantity(1, "m")
    >>> dc.fields(q)
    (Field(name='value',...), Field(name='unit',...))

    """
    value_field = dataclasses.Field(
        dataclasses.MISSING,
        dataclasses.MISSING,
        True,  # noqa: FBT003
        False,  # noqa: FBT003
        False,  # noqa: FBT003
        False,  # noqa: FBT003
        {},
        False,  # noqa: FBT003
    )
    value_field.name = "value"
    unit_field = dataclasses.Field(
        dataclasses.MISSING,
        dataclasses.MISSING,
        True,  # noqa: FBT003
        False,  # noqa: FBT003
        False,  # noqa: FBT003
        False,  # noqa: FBT003
        {},
        False,  # noqa: FBT003
    )
    unit_field.name = "unit"
    return (value_field, unit_field)



"""Units objects in JAX.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__ = ("unit", "unit_of", "AbstractUnit")

from typing import Any, TypeAlias

import astropy.units as apyu
from plum import dispatch

import unxt_api as uapi
from unxt.dims import AbstractDimension

AbstractUnit: TypeAlias = apyu.Unit | apyu.UnitBase | apyu.CompositeUnit


# ===================================================================
# Construct units


@dispatch
def unit(obj: AbstractUnit, /) -> AbstractUnit:
    """Construct the units from a units object.

    Examples
    --------
    >>> import unxt as u
    >>> m = u.unit("m")

    >>> u.unit(m) is m
    True

    """
    return obj


@dispatch
def unit(obj: str, /) -> AbstractUnit:
    """Construct units from a string.

    Examples
    --------
    >>> import unxt as u
    >>> m = u.unit("m")
    >>> m
    Unit("m")

    """
    return apyu.Unit(obj)


# ===================================================================
# Get units


@dispatch
def unit_of(obj: Any, /) -> None:
    """Return the units of an object.

    Examples
    --------
    >>> import unxt as u
    >>> print(u.unit_of(1))
    None

    """
    return None  # noqa: RET501


@dispatch
def unit_of(obj: AbstractUnit, /) -> AbstractUnit:
    """Return the units of an unit.

    Examples
    --------
    >>> import unxt as u
    >>> m = u.unit("m")

    >>> u.unit_of(m)
    Unit("m")

    """
    return obj


# ===================================================================
# Get dimensions


@dispatch
def dimension_of(obj: AbstractUnit, /) -> AbstractDimension:
    """Return the dimensions of the given units.

    Examples
    --------
    >>> import unxt as u
    >>> u.dimension_of(u.unit("km"))
    PhysicalType('length')

    """
    return uapi.dimension(obj.physical_type)



"""Unit tests for unit systems API (unxt_api._unitsystems)."""

import plum

import unxt_api as uapi

# ==============================================================================
# Tests for unitsystem_of()
# ==============================================================================


def test_unitsystem_of_is_abstract_dispatch() -> None:
    """Test that unitsystem_of is an abstract dispatch function."""
    assert isinstance(uapi.unitsystem_of, plum.function.Function)
    assert hasattr(uapi.unitsystem_of, "methods")


def test_unitsystem_of_accepts_any_type() -> None:
    """Test that unitsystem_of accepts Any type in signature."""
    assert "unitsystem_of" in dir(uapi)


def test_unitsystem_of_no_default_implementation() -> None:
    """Test that calling unitsystem_of without implementation may return a default.

    Note: When unxt is imported (e.g., during doctest execution), it may
    register a default implementation that returns a system for unknown types.
    """

    class NoDispatchType:
        pass

    obj = NoDispatchType()

    try:
        result = uapi.unitsystem_of(obj)
        # If it doesn't raise, unxt has provided a default implementation
        # (typically returns DimensionlessUnitSystem or None)
        assert result is not None or result is None
    except plum.resolver.NotFoundLookupError:
        # This is also acceptable if unxt hasn't been imported
        pass


def test_unitsystem_of_can_register_custom_dispatch(custom_quantity_type) -> None:
    """Test that custom dispatches can be registered."""

    class MockUnitSystem:
        def __init__(self, name: str) -> None:
            self.name = name

    @plum.dispatch
    def unitsystem_of(obj: custom_quantity_type, /) -> MockUnitSystem:
        return MockUnitSystem("SI")

    quantity = custom_quantity_type(42, "m")
    result = unitsystem_of(quantity)

    assert isinstance(result, MockUnitSystem)
    assert result.name == "SI"


def test_unitsystem_of_multiple_dispatches_possible() -> None:
    """Test that multiple dispatches can coexist."""
    initial_count = len(uapi.unitsystem_of.methods)
    assert initial_count >= 0


# ==============================================================================
# API consistency tests
# ==============================================================================


def test_unitsystem_of_is_exported() -> None:
    """Test that unitsystem_of is exported from unxt_api."""
    assert hasattr(uapi, "unitsystem_of")


def test_unitsystem_of_in_all() -> None:
    """Test that unitsystem_of is in __all__."""
    assert "unitsystem_of" in uapi.__all__


def test_unitsystem_of_is_dispatch_function() -> None:
    """Test that unitsystem_of is a dispatch function."""
    assert isinstance(uapi.unitsystem_of, plum.function.Function)



# pylint: disable=import-error, no-member, unsubscriptable-object
#    b/c it doesn't understand dataclass fields

__all__ = ("AbstractQuantity", "is_any_quantity")

import functools as ft
import warnings
from collections.abc import Mapping
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    NoReturn,
    TypeAlias,
    TypeGuard,
    cast,
)
from typing_extensions import override

import equinox as eqx
import jax
import jax.core
import numpy as np
import quax_blocks
import wadler_lindig as wl
from jax._src.numpy.array_methods import _IndexUpdateHelper, _IndexUpdateRef
from jaxtyping import Array, ArrayLike, Bool, ScalarLike, Shaped
from plum import add_promotion_rule, dispatch, type_nonparametric
from quax import ArrayValue

import quaxed.numpy as jnp
from dataclassish import field_items, replace

import unxt_api as uapi
from .mixins import AstropyQuantityCompatMixin, IPythonReprMixin, NumPyCompatMixin
from .value import StaticValue
from unxt.units import AbstractUnit

if TYPE_CHECKING:
    from typing import Self


ArrayLikeSequence: TypeAlias = list[ScalarLike] | tuple[ScalarLike, ...]


class AbstractQuantity(
    AstropyQuantityCompatMixin,
    NumPyCompatMixin,
    IPythonReprMixin,
    ArrayValue,
    quax_blocks.NumpyBinaryOpsMixin[Any, "AbstractQuantity"],
    quax_blocks.NumpyComparisonMixin[Any, Bool[Array, "*shape"]],  # TODO: shape hint
    quax_blocks.NumpyUnaryMixin["AbstractQuantity"],
    quax_blocks.NumpyRoundMixin["AbstractQuantity"],
    quax_blocks.NumpyTruncMixin["AbstractQuantity"],
    quax_blocks.NumpyFloorMixin["AbstractQuantity"],
    quax_blocks.NumpyCeilMixin["AbstractQuantity"],
    quax_blocks.LaxLenMixin,
    quax_blocks.LaxLengthHintMixin,
):
    """Represents a Quantity with a unit.

    Attributes
    ----------
    short_name : str | None
        Optional short name for the class used in wadler-lindig printing when
        ``use_short_name=True``. Defaults to ``None``.

    Examples
    --------
    >>> import unxt as u

    From an integer:

    >>> u.Quantity(1, "m")
    Quantity(Array(1, dtype=int32, ...), unit='m')

    From a float:

    >>> u.Quantity(1.0, "m")
    Quantity(Array(1., dtype=float32, ...), unit='m')

    From a list:

    >>> u.Quantity([1, 2, 3], "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    From a tuple:

    >>> u.Quantity((1, 2, 3), "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    From a `numpy.ndarray`:

    >>> import numpy as np
    >>> u.Quantity(np.array([1, 2, 3]), "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    From a `jax.Array`:

    >>> import jax.numpy as jnp
    >>> u.Quantity(jnp.array([1, 2, 3]), "m")
    Quantity(Array([1, 2, 3], dtype=int32), unit='m')

    The unit can also be given as a `astropy.units.Unit`:

    >>> import astropy.units as apyu
    >>> u.Quantity(1, apyu.m)
    Quantity(Array(1, dtype=int32, ...), unit='m')

    """

    value: eqx.AbstractVar[Shaped[Array | StaticValue, "*shape"]]
    """The value of the `AbstractQuantity`."""

    unit: eqx.AbstractVar[AbstractUnit]
    """The unit associated with this value."""

    # ---------------------------------------------------------------
    # Constructors

    @classmethod
    @dispatch.abstract
    def from_(
        cls: "type[AbstractQuantity]",
        *args: Any,
        **kwargs: Any,
    ) -> "AbstractQuantity":
        raise NotImplementedError  # pragma: no cover

    # See below for additional constructors

    # ===============================================================
    # Quantity API

    def uconvert(self, u: Any, /) -> "AbstractQuantity":
        """Convert the quantity to the given units.

        See Also
        --------
        `unxt.uconvert` : convert a quantity to a new unit.

        Examples
        --------
        >>> import unxt as u

        >>> q = u.Quantity(1, "m")
        >>> q.uconvert("cm")
        Quantity(Array(100., dtype=float32, ...), unit='cm')

        """
        return uapi.uconvert(u, self)

    def ustrip(self, u: Any, /) -> Array:
        """Return the value in the given units.

        See Also
        --------
        `unxt.ustrip` : strip the units from a quantity.

        Examples
        --------
        >>> import unxt as u

        >>> q = u.Quantity(1, "m")
        >>> q.ustrip("cm")
        Array(100., dtype=float32, weak_type=True)

        """
        return uapi.ustrip(u, self)

    # ===============================================================
    # Quax API

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of the array."""
        return self.value.shape

    def materialise(self) -> NoReturn:
        msg = f"Refusing to materialise `{type(self).__name__}`."
        raise RuntimeError(msg)

    def aval(self) -> jax.core.ShapedArray:
        value = self.value
        if isinstance(value, StaticValue):
            value = jnp.asarray(value.array)
        return jax.typeof(value)

    def enable_materialise(self, _: bool = True) -> "Self":  # noqa: FBT001, FBT002
        return replace(self, value=self.value, unit=self.unit)

    # ===============================================================
    # Plum API

    #: This tells `plum` that this type can be efficiently cached.
    __faithful__: ClassVar[bool] = True

    # ===============================================================
    # Array API

    def __array_namespace__(self, *, api_version: Any = None) -> ModuleType:
        """Return the namespace for the array API.

        Here we return the `quaxed.numpy` module, which is a drop-in replacement
        for `jax.numpy`, but allows for array-ish objects to be used in place of
        `jax` arrays. See `quax` for more information.

        """
        return jnp  # quaxed.numpy

    # ---------------------------------------------------------------
    # attributes

    @property
    def dtype(self) -> jax.numpy.dtype:
        """Data type of the array.

        Examples
        --------
        >>> import unxt as u
        >>> u.Quantity(1, "m").dtype
        dtype('int32')

        """
        return self.value.dtype

    @property
    def device(self) -> jax.Device:
        """Device where the array is located.

        Examples
        --------
        >>> import unxt as u
        >>> u.Quantity(1, "m").device
        CpuDevice(id=0)

        """
        return self.value.devices().pop()

    @property
    def mT(self) -> "AbstractQuantity":  # noqa: N802
        """Matrix transpose of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([[0, 1], [1, 2]], "m")
        >>> q.mT
        Quantity(Array([[0, 1],
                                  [1, 2]], dtype=int32), unit='m')

        """
        return replace(self, value=jnp.matrix_transpose(self.value))

    @property
    def ndim(self) -> int:
        """Number of dimensions.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([[1]], "m")
        >>> q.ndim
        2

        """
        return self.value.ndim

    @property
    def size(self) -> int:
        """Total number of elements.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.size
        3

        """
        return self.value.size

    @property
    def T(self) -> "AbstractQuantity":  # noqa: N802
        """Transpose of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([[0, 1], [1, 2]], "m")
        >>> q.T
        Quantity(Array([[0, 1],
                                  [1, 2]], dtype=int32), unit='m')

        """
        return replace(self, value=self.value.T)

    # required to override mixin methods
    __eq__ = quax_blocks.NumpyEqMixin.__eq__  # type: ignore[assignment,unused-ignore]

    # ---------------------------------------------------------------
    # methods

    def __bool__(self) -> bool:
        """Convert a zero-dimensional array to a Python bool object.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([0, 1], "m")

        >>> bool(q[0])
        False

        >>> bool(q[1])
        True

        """
        return bool(self.value)

    def __complex__(self) -> complex:
        """Convert the array to a Python complex object.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "")
        >>> complex(q)
        (1+0j)

        """
        return complex(uapi.ustrip("", self))

    def __dlpack__(self, *args: Any, **kwargs: Any) -> Any:
        """Export the array for consumption as a DLPack capsule."""
        raise NotImplementedError

    def __dlpack_device__(self, *args: Any, **kwargs: Any) -> Any:
        """Return device type and device ID in DLPack format."""
        raise NotImplementedError

    def __float__(self) -> float:
        """Convert the array to a Python float object.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "")
        >>> float(q)
        1.0

        """
        return float(uapi.ustrip("", self))

    def __getitem__(self, key: Any) -> "AbstractQuantity":
        """Get an item from the array.

        This is a simple wrapper around the `__getitem__` method of the array,
        calling `replace` to only update the value.

        """
        return replace(self, value=self.value[key])

    def __index__(self) -> int:
        """Convert a zero-dimensional integer array to a Python int object.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "")
        >>> q.__index__()
        1

        """
        return uapi.ustrip("", self).__index__()

    def __int__(self) -> int:
        """Convert a zero-dimensional array to a Python int object.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "")
        >>> int(q)
        1

        """
        return int(uapi.ustrip("", self))

    def __setitem__(self, key: Any, value: Any) -> None:
        """Set an item in the array.

        This is a simple wrapper around the `__setitem__` method of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> try:
        ...     q[0] = 2
        ... except Exception as e:
        ...     print("jax arrays do not support in-place item assignment")
        jax arrays do not support in-place item assignment

        """
        self.value[key] = value

    def to_device(self, device: None | jax.Device = None) -> "AbstractQuantity":
        """Move the array to a new device.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "m")
        >>> q.to_device(None)
        Quantity(Array(1, dtype=int32, weak_type=True), unit='m')

        """
        return replace(self, value=self.value.to_device(device))

    # ===============================================================
    # JAX API

    def __iter__(self) -> Any:
        """Iterate over the Quantity's value.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> [x for x in q]
        [Quantity(Array(1, dtype=int32), unit='m'),
         Quantity(Array(2, dtype=int32), unit='m'),
         Quantity(Array(3, dtype=int32), unit='m')]

        """
        yield from (self[i] for i in range(len(self.value)))

    def argmax(self, *args: Any, **kwargs: Any) -> Array:
        """Return the indices of the maximum value.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.argmax()
        Array(2, dtype=int32)

        """
        return self.value.argmax(*args, **kwargs)

    def argmin(self, *args: Any, **kwargs: Any) -> Array:
        """Return the indices of the minimum value.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.argmin()
        Array(0, dtype=int32)

        """
        return self.value.argmin(*args, **kwargs)

    def astype(self, *args: Any, **kwargs: Any) -> "AbstractQuantity":
        """Copy the array and cast to a specified dtype.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.dtype
        dtype('int32')

        >>> q.astype(float)
        Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

        """
        return replace(self, value=self.value.astype(*args, **kwargs))

    @ft.partial(property, doc=jax.Array.at.__doc__)
    def at(self) -> "_QuantityIndexUpdateHelper":
        return _QuantityIndexUpdateHelper(self)  # type: ignore[arg-type]

    def block_until_ready(self) -> "AbstractQuantity":
        """Block until the array is ready.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "m")
        >>> q.block_until_ready() is q
        True

        """
        _ = self.value.block_until_ready()
        return self

    def devices(self) -> set[jax.Device]:
        """Return the devices where the array is located.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "m")
        >>> q.devices()
        {CpuDevice(id=0)}

        """
        return self.value.devices()

    def flatten(self) -> "AbstractQuantity":
        """Return a flattened version of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([[1, 2], [3, 4]], "m")
        >>> q.flatten()
        Quantity(Array([1, 2, 3, 4], dtype=int32), unit='m')

        """
        return replace(self, value=self.value.flatten())

    def max(self, *args: Any, **kwargs: Any) -> "AbstractQuantity":
        """Return the maximum value.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.max()
        Quantity(Array(3, dtype=int32), unit='m')

        """
        return replace(self, value=self.value.max(*args, **kwargs))

    def mean(self, *args: Any, **kwargs: Any) -> "AbstractQuantity":
        """Return the mean value.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.mean()
        Quantity(Array(2., dtype=float32), unit='m')

        """
        return replace(self, value=self.value.mean(*args, **kwargs))

    def min(self, *args: Any, **kwargs: Any) -> "AbstractQuantity":
        """Return the minimum value.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.min()
        Quantity(Array(1, dtype=int32), unit='m')

        """
        return replace(self, value=self.value.min(*args, **kwargs))

    def ravel(self) -> "AbstractQuantity":
        """Return a flattened version of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([[1, 2], [3, 4]], "m")
        >>> q.ravel()
        Quantity(Array([1, 2, 3, 4], dtype=int32), unit='m')

        """
        return replace(self, value=self.value.ravel())

    def reshape(self, *args: Any, order: str = "C") -> "AbstractQuantity":
        """Return a reshaped version of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3, 4], "m")
        >>> q.reshape(2, 2)
        Quantity(Array([[1, 2],
                                  [3, 4]], dtype=int32), unit='m')

        """
        __tracebackhide__ = True  # pylint: disable=unused-variable
        return replace(self, value=self.value.reshape(*args, order=order))

    def round(self, *args: Any, **kwargs: Any) -> "AbstractQuantity":
        """Round the array to the given number of decimals.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1.1, 2.2, 3.3], "m")
        >>> q.round(0)
        Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

        """
        return replace(self, value=self.value.round(*args, **kwargs))

    @property
    def sharding(self) -> Any:
        """Return the sharding configuration of the array.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3], "m")
        >>> q.sharding
        SingleDeviceSharding(device=..., memory_kind=...)

        """
        return self.value.sharding

    def squeeze(self, *args: Any, **kwargs: Any) -> "AbstractQuantity":
        """Return the array with all single-dimensional entries removed.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([[[1], [2], [3]]], "m")
        >>> q.squeeze()
        Quantity(Array([1, 2, 3], dtype=int32), unit='m')

        """
        return replace(self, value=self.value.squeeze(*args, **kwargs))

    # ===============================================================
    # Python stuff

    def __hash__(self) -> int:
        """Return the hash of the quantity.

        This is the hash of the value and the unit, however since the value is
        generally not hashable this will generally raise an exception.
        Defining the `__hash__` method is required for the `AbstractQuantity` to
        be considered immutable, e.g. by `dataclasses.dataclass`.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity(1, "m")
        >>> try:
        ...     hash(q)
        ... except TypeError as e:
        ...     print(e)
        unhashable type: 'jaxlib...ArrayImpl'

        """
        return hash((self.value, self.unit))

    def __pdoc__(
        self,
        *,
        named_unit: bool = True,
        short_arrays: bool | Literal["compact"] = True,
        use_short_name: bool = False,
        **kwargs: Any,
    ) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation of this class.

        This is used for the `__repr__` and `__str__` methods or when using the
        `wadler_lindig` library.

        Parameters
        ----------
        named_unit
            If `True`, the unit is included in the representation as a named
            argument. If `False`, the unit is included as a positional argument.
            For example, ``Quantity(<array>, unit='m')`` versus
            ``Quantity(<array>, 'm')``.
        short_arrays
            If 'compact', the array is printed in a compact form,
            ``Quantity(Array([1., 2.], dtype=float32), unit='')`` to
            ``Quantity([1., 2.], unit='')``. If `True` or `False` it uses the
            default `wadler_lindig` behavior.
        use_short_name
            If `True` and the class has a ``short_name`` class variable, use the
            short name instead of the full class name. For example, ``Q(...)``
            instead of ``Quantity(...)``.
        kwargs
            Additional keyword arguments ``wadler_lindig.pdoc`` method for
            formatting the value, stringified unit, and other fields.

        Examples
        --------
        >>> import unxt as u
        >>> import wadler_lindig as wl

        >>> q = u.quantity.BareQuantity([1, 2, 3], "m")

        The default pretty printing:

        >>> wl.pprint(q)
        BareQuantity(i32[3], unit='m')

        The `str` method uses this as well:

        >>> print(q)
        BareQuantity(i32[3], unit='m')

        Arrays can be printed in full:

        >>> wl.pprint(q, short_arrays=False)
        BareQuantity(Array([1, 2, 3], dtype=int32), unit='m')

        The `repr` method uses this setting:

        >>> print(repr(q))
        BareQuantity(Array([1, 2, 3], dtype=int32), unit='m')

        The units can be turned from a named argument to a positional argument
        by setting `named_unit=False`:

        >>> wl.pprint(q, named_unit=False)
        BareQuantity(i32[3], 'm')

        The arrays can be printed in a compact but informative form by setting
        ``short_arrays="compact"``:

        >>> wl.pprint(q, short_arrays="compact")
        BareQuantity([1, 2, 3], unit='m')

        The class short name can be used if available:

        >>> wl.pprint(q, use_short_name=True)
        BareQuantity(i32[3], unit='m')

        """
        # Class Name
        if use_short_name and getattr(self, "short_name", None) is not None:
            cls_name = wl.TextDoc(self.short_name)
        else:
            cls_name = wl.TextDoc(type_nonparametric(self).__name__)

        # Object fields
        fs = dict(field_items(self))
        del fs["value"]
        del fs["unit"]

        # Customize value representation
        # (backward compatibility for `compact_arrays` argument)
        # TODO: remove in v2.0
        if "compact_arrays" in kwargs:
            short_arrays = kwargs.pop("compact_arrays")
            warnings.warn(
                "`compact_arrays` argument is deprecated; use "
                "`short_arrays='compact'` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        match short_arrays:
            case "compact":
                kwargs["custom"] = custom_pdoc_noarray
            case True:
                kwargs["custom"] = custom_pdoc_no_kind
                kwargs["short_arrays"] = True
            case False:
                kwargs["short_arrays"] = False

        # Make the pdocs for the base and extra fields.
        base_fields = [
            wl.pdoc(self.value, **kwargs),
            wl.TextDoc("unit=" if named_unit else "")
            + wl.pdoc(self.unit.to_string(), **kwargs),
        ]
        extra_fields = wl.named_objs(tuple(fs.items()), **kwargs)

        # Construct and return the Wadler-Lindig document.
        return (
            cls_name
            + wl.TextDoc("(")
            + wl.join(wl.comma, base_fields + extra_fields)
            .group()
            .nest(kwargs.get("indent", 4))
            + wl.TextDoc(")")
        )

    def __repr__(self) -> str:
        # TODO: make named_unit False?
        return wl.pformat(self, short_arrays=False, named_unit=True, indent=4)

    def __str__(self) -> str:
        # TODO: make named_unit False?
        return wl.pformat(self, short_arrays=True, named_unit=True, indent=4)


# -----------------------------------------------
# Register additional constructors


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity],
    value: ArrayLike | ArrayLikeSequence,
    unit: Any,
    /,
    *,
    dtype: Any = None,
) -> AbstractQuantity:
    """Construct a `unxt.Quantity` from an array-like value and a unit.

    :param value: The array-like value.
    :param unit: The unit of the value.
    :param dtype: The data type of the array (keyword-only).

    Examples
    --------
    For this example we'll use the `Quantity` class. The same applies to
    any subclass of `AbstractQuantity`.

    >>> import jax.numpy as jnp
    >>> import unxt as u

    >>> x = jnp.array([1.0, 2, 3])
    >>> u.Quantity.from_(x, "m")
    Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

    >>> u.Quantity.from_([1.0, 2, 3], "m")
    Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

    >>> u.Quantity.from_((1.0, 2, 3), "m")
    Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

    """
    # Dispatch on both arguments.
    # Construct using the standard `__init__` method.
    return cls(jnp.asarray(value, dtype=dtype), unit)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity],
    value: ArrayLike | ArrayLikeSequence,
    /,
    *,
    unit: Any,
    dtype: Any = None,
) -> AbstractQuantity:
    """Make a `unxt.AbstractQuantity` from an array-like value and a unit kwarg.

    Examples
    --------
    For this example we'll use the `unxt.Quantity` class. The same applies
    to any subclass of `unxt.AbstractQuantity`.

    >>> import unxt as u
    >>> u.Quantity.from_([1.0, 2, 3], unit="m")
    Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

    """
    # Dispatch on the `value` only. Dispatch to the full constructor.
    return cls.from_(value, unit, dtype=dtype)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity], *, value: Any, unit: Any, dtype: Any = None
) -> AbstractQuantity:
    """Construct a `AbstractQuantity` from value and unit kwargs.

    Examples
    --------
    For this example we'll use the `Quantity` class. The same applies to
    any subclass of `AbstractQuantity`.

    >>> import unxt as u
    >>> u.Quantity.from_(value=[1.0, 2, 3], unit="m")
    Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

    """
    # Dispatched on no argument. Dispatch to the full constructor.
    return cls.from_(value, unit, dtype=dtype)


@AbstractQuantity.from_.dispatch
def from_(cls: type[AbstractQuantity], mapping: Mapping[str, Any]) -> AbstractQuantity:
    """Construct a `Quantity` from a Mapping.

    Examples
    --------
    For this example we'll use the `Quantity` class. The same applies to
    any subclass of `AbstractQuantity`.

    >>> import jax.numpy as jnp
    >>> import unxt as u

    >>> x = jnp.array([1.0, 2, 3])
    >>> q = u.Quantity.from_({"value": x, "unit": "m"})
    >>> q
    Quantity(Array([1., 2., 3.], dtype=float32), unit='m')

    >>> u.Quantity.from_({"value": q, "unit": "km"})
    Quantity(Array([0.001, 0.002, 0.003], dtype=float32), unit='km')

    """
    # Dispatch on both arguments.
    # Construct using the standard `__init__` method.
    return cls.from_(**mapping)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity],
    value: AbstractQuantity,
    unit: Any,
    /,
    *,
    dtype: Any = None,
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1, "m")
    >>> u.Quantity.from_(q, "cm")
    Quantity(Array(100., dtype=float32, ...), unit='cm')

    """
    value = jnp.asarray(uapi.uconvert(unit, value), dtype=dtype)
    return cls(uapi.ustrip(unit, value), unit)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity],
    value: AbstractQuantity,
    unit: None,
    /,
    *,
    dtype: Any = None,
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`.

    The `value` is converted to the new `unit`.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1, "m")
    >>> u.Quantity.from_(q, None)
    Quantity(Array(1, dtype=int32, ...), unit='m')

    """
    value = jnp.asarray(value, dtype=dtype)
    unit = uapi.unit_of(value)
    return cls(uapi.ustrip(unit, value), unit)


@AbstractQuantity.from_.dispatch
def from_(
    cls: type[AbstractQuantity],
    value: AbstractQuantity,
    /,
    *,
    unit: Any | None = None,
    dtype: Any = None,
) -> AbstractQuantity:
    """Construct a `Quantity` from another `Quantity`, with no unit change."""
    unit = value.unit if unit is None else unit
    value = jnp.asarray(uapi.uconvert(unit, value), dtype=dtype)
    return cls(uapi.ustrip(unit, value), unit)


# -----------------------------------------------
# Promotion rules

add_promotion_rule(AbstractQuantity, AbstractQuantity, AbstractQuantity)


# ===============================================================
# Support for ``at``.


# `_QuantityIndexUpdateHelper` is defined up here because it is used in the
# runtime-checkable type annotation in `AbstractQuantity.at`.
# `_QuantityIndexUpdateRef` is defined after `AbstractQuantity` because it
# references `AbstractQuantity` in its runtime-checkable type annotations.
class _QuantityIndexUpdateHelper(_IndexUpdateHelper):
    def __getitem__(self, index: Any) -> "_IndexUpdateRef":
        return _QuantityIndexUpdateRef(self.array, index)

    def __repr__(self) -> str:
        """Return a string representation of the object.

        Examples
        --------
        >>> import unxt as u
        >>> q = u.Quantity([1, 2, 3, 4], "m")
        >>> q.at
        _QuantityIndexUpdateHelper(Quantity(Array([1, 2, 3, 4], dtype=int32), unit='m'))

        """
        return f"_QuantityIndexUpdateHelper({self.array!r})"


class _QuantityIndexUpdateRef(_IndexUpdateRef):
    # This is a subclass of `_IndexUpdateRef` that is used to implement the `at`
    # attribute of `AbstractQuantity`. See also `_QuantityIndexUpdateHelper`.

    array: AbstractQuantity  # type: ignore[assignment]

    def __repr__(self) -> str:
        return super().__repr__().replace("_IndexUpdateRef", "_QuantityIndexUpdateRef")

    @override
    def get(  # type: ignore[override]
        self, *, fill_value: AbstractQuantity | None = None, **kw: Any
    ) -> AbstractQuantity:
        # TODO: by quaxified super
        value = self.array.value.at[self.index].get(
            fill_value=(
                fill_value
                if fill_value is None
                else uapi.ustrip(self.array.unit, fill_value)
            ),
            **kw,
        )
        return replace(self.array, value=value)

    @override
    def set(self, values: AbstractQuantity, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        # TODO: by quaxified super
        value = self.array.value.at[self.index].set(
            uapi.ustrip(self.array.unit, values), **kw
        )
        return replace(self.array, value=value)

    @override
    def apply(self, func: Any, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        raise NotImplementedError  # TODO: by quaxified super

    @override
    def add(self, values: AbstractQuantity, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        # TODO: by quaxified super
        value = self.array.value.at[self.index].add(
            uapi.ustrip(self.array.unit, values), **kw
        )
        return replace(self.array, value=value)

    @override
    def multiply(self, values: ArrayLike, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        values = eqx.error_if(
            values, isinstance(values, AbstractQuantity), "values cannot be a Quantity"
        )  # TODO: can permit dimensionless quantities.

        # TODO: by quaxified super
        value = self.array.value.at[self.index].multiply(values, **kw)
        return replace(self.array, value=value)

    mul = multiply  # type: ignore[assignment]

    @override
    def divide(self, values: ArrayLike, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        values = eqx.error_if(
            values, isinstance(values, AbstractQuantity), "values cannot be a Quantity"
        )  # TODO: can permit dimensionless quantities.

        # TODO: by quaxified super
        value = self.array.value.at[self.index].divide(values, **kw)
        return replace(self.array, value=value)

    @override
    def power(self, values: ArrayLike, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        raise NotImplementedError

    @override
    def min(self, values: AbstractQuantity, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        # TODO: by quaxified super
        value = self.array.value.at[self.index].min(
            uapi.ustrip(self.array.unit, values), **kw
        )
        return replace(self.array, value=value)

    @override
    def max(self, values: AbstractQuantity, **kw: Any) -> AbstractQuantity:  # type: ignore[override]
        # TODO: by quaxified super
        value = self.array.value.at[self.index].max(
            uapi.ustrip(self.array.unit, values), **kw
        )
        return replace(self.array, value=value)


# This is public!
def is_any_quantity(obj: Any, /) -> TypeGuard[AbstractQuantity]:
    """Check if an object is an instance of `unxt.quantity.AbstractQuantity`.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Quantity(1, "m")
    >>> is_any_quantity(q)
    True

    """
    return isinstance(obj, AbstractQuantity)


# TODO: replace with `equinox.internal.TreeWLCustom` when available.
def custom_pdoc_no_kind(obj: Any) -> wl.AbstractDoc | None:
    """Return custom pdoc for ``AbstractQuantity`` objects."""
    if isinstance(obj, jax.Array):
        dtype = obj.dtype.name
        # Added in JAX 0.4.32 to `ShapeDtypeStruct`
        if getattr(obj, "weak_type", False):
            dtype = f"weak_{dtype}"
        return wl.array_summary(obj.shape, dtype, kind=None)
    return None


def custom_pdoc_noarray(obj: Any) -> wl.AbstractDoc | None:
    """Return custom pdoc for ``AbstractQuantity`` objects."""
    if isinstance(obj, jax.Array):
        return wl.TextDoc(np.array2string(cast("np.ndarray", obj), separator=", "))
    return None


@StaticValue.from_.dispatch
def from_(cls: StaticValue, value: AbstractQuantity, /) -> StaticValue:
    """Disallow conversion of `AbstractQuantity` to a value."""
    msg = (
        f"Cannot convert '{type(value).__name__}' to a value. "
        "For a Quantity, use the `.from_` constructor instead."
    )
    raise TypeError(msg)


@dispatch
def convert_to_quantity_value(obj: AbstractQuantity, /) -> NoReturn:
    """Disallow conversion of `AbstractQuantity` to a value.

    >>> import unxt as u
    >>> from unxt.quantity import convert_to_quantity_value

    >>> try:
    ...     convert_to_quantity_value(u.Quantity(1, "m"))
    ... except TypeError as e:
    ...     print(e)
    Cannot convert 'Quantity[PhysicalType('length')]' to a value.
    For a Quantity, use the `.from_` constructor instead.

    """
    msg = (
        f"Cannot convert '{type(obj).__name__}' to a value. "
        "For a Quantity, use the `.from_` constructor instead."
    )
    raise TypeError(msg)



"""Hypothesis strategies for named dimensions.

Derived from Astropy physical types list at
https://github.com/astropy/astropy/blob/dc7afddeb21316baa9f5d71f6d32c35fd02ca3db/astropy/units/physical.py.
The names are captured here as a finalized tuple and filtered to those that
`unxt.dimension` can resolve.
"""

__all__ = ("DIMENSION_NAMES", "named_dimensions")

from typing import Final

from hypothesis import strategies as st

import unxt as u

# The Astropy physical type names that map cleanly to unxt dimensions.
DIMENSION_NAMES: Final[tuple[str, ...]] = (
    "absement",
    "absity",
    "acceleration",
    "action",
    "amount of substance",
    "angle",
    "angular acceleration",
    "angular frequency",
    "angular momentum",
    "angular speed",
    "angular velocity",
    "area",
    "bandwidth",
    "catalytic activity",
    "chemical potential",
    "column density",
    "compressibility",
    "crackle",
    "data quantity",
    "diffusivity",
    "dimensionless",
    "dose of ionizing radiation",
    "dynamic viscosity",
    "electrical capacitance",
    "electrical charge",
    "electrical charge density",
    "electrical conductance",
    "electrical conductivity",
    "electrical current",
    "electrical current density",
    "electrical dipole moment",
    "electrical field strength",
    "electrical flux density",
    "electrical impedance",
    "electrical mobility",
    "electrical potential",
    "electrical reactance",
    "electrical resistance",
    "electrical resistivity",
    "electromagnetic field strength",
    "electron density",
    "electron flux",
    "energy",
    "energy density",
    "energy flux",
    "entropy",
    "force",
    "frequency",
    "frequency drift",
    "heat capacity",
    "illuminance",
    "impulse",
    "inductance",
    "irradiance",
    "jerk",
    "jolt",
    "jounce",
    "kinematic viscosity",
    "length",
    "linear density",
    "luminance",
    "luminous efficacy",
    "luminous emittance",
    "luminous flux",
    "luminous intensity",
    "magnetic field strength",
    "magnetic flux",
    "magnetic flux density",
    "magnetic helicity",
    "magnetic moment",
    "magnetic reluctance",
    "mass",
    "mass attenuation coefficient",
    "mass density",
    "mass flux",
    "molality",
    "molar concentration",
    "molar conductivity",
    "molar heat capacity",
    "molar volume",
    "moment of inertia",
    "momentum",
    "momentum density",
    "number density",
    "opacity",
    "particle flux",
    "permeability",
    "permittivity",
    "photon flux",
    "photon flux density",
    "photon flux density wav",
    "photon surface brightness",
    "photon surface brightness wav",
    "plate scale",
    "polarization density",
    "pop",
    "pounce",
    "power",
    "power density",
    "pressure",
    "radiance",
    "radiant flux",
    "radiant intensity",
    "reaction rate",
    "snap",
    "solid angle",
    "specific energy",
    "specific entropy",
    "specific heat capacity",
    "specific volume",
    "spectral flux density",
    "spectral flux density wav",
    "speed",
    "stress",
    "surface brightness",
    "surface brightness wav",
    "surface charge density",
    "surface mass density",
    "surface tension",
    "temperature",
    "temperature gradient",
    "thermal conductance",
    "thermal conductivity",
    "thermal resistance",
    "thermal resistivity",
    "time",
    "torque",
    "velocity",
    "volume",
    "volumetric flow rate",
    "volumetric rate",
    "wavenumber",
    "work",
    "yank",
)


@st.composite
def named_dimensions(draw: st.DrawFn) -> u.AbstractDimension:
    """Generate a named dimension from Astropy's physical type catalogue.

    This strategy samples from a finalized, pre-computed set of 134 named
    physical dimensions derived from Astropy's physical type system. These
    dimensions are guaranteed to be resolvable by `unxt.dimension()`.

    The set of dimension names is curated from Astropy's physical types and
    excludes names that cannot be directly interpreted by unxt (e.g., names
    with parenthetical qualifiers like "electrical charge (EMU)").

    Parameters
    ----------
    draw : st.DrawFn
        Hypothesis draw function (automatically provided by @st.composite).

    Returns
    -------
    unxt.AbstractDimension
        A randomly selected dimension from the set of named dimensions.

    Examples
    --------
    Basic usage - generate a dimension:

    >>> from hypothesis import given
    >>> import unxt as u
    >>> import unxt_hypothesis as ust

    >>> @given(dim=ust.named_dimensions())
    ... def test_physics_dimension(dim):
    ...     assert isinstance(dim, u.AbstractDimension)

    Using with unit generation to create quantities of any dimension:

    >>> @given(
    ...     dim=ust.named_dimensions(),
    ...     q=ust.quantities(u.dimension("length"), shape=3),
    ... )
    ... def test_dimension_and_quantity(dim, q):
    ...     # dim varies across test runs from 134 physical types
    ...     assert isinstance(dim, u.AbstractDimension)
    ...     assert q.shape == (3,)

    Combine with units strategy:

    >>> @given(unit=ust.units(ust.named_dimensions()))
    ... def test_any_unit(unit):
    ...     # Generate units from any named physical dimension
    ...     assert isinstance(unit, u.AbstractUnit)
    ...     assert u.dimension_of(unit) in [
    ...         u.dimension(name) for name in ust.DIMENSION_NAMES
    ...     ]

    Property-based testing across all 134 physical types:

    >>> @given(dim=ust.named_dimensions())
    ... def test_all_dimensions_are_valid(dim):
    ...     # This test will run with 100 different random dimensions
    ...     # (by default, Hypothesis runs 100 examples)
    ...     assert isinstance(dim, u.AbstractDimension)
    ...     assert dim in {u.dimension(name) for name in ust.DIMENSION_NAMES}

    See Also
    --------
    DIMENSION_NAMES : The tuple of all available dimension names.
    unxt.dimension : Create a dimension from a name string.
    units : Generate units with a specific dimension.
    quantities : Generate quantities with a specific dimension.

    """
    name = draw(st.sampled_from(DIMENSION_NAMES))
    return u.dimension(name)
